/*
 *  gcube
 *  Copyright (c) 2004 monk
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *
 *  GX
 *  will be split into few smaller files
 *         
 */

#include "hw_gx.h"
#include <SDL/SDL_opengl.h>

#ifdef LIL_ENDIAN
# define MASK_ALPHA		0xff000000
#else
# define MASK_ALPHA		0xff
#endif

// check only main fifo, not pregenerated display lists
#define EXCEEDES_LIST_BOUNDARY(m,s)		((m >= CP_FIFO_END) ? FALSE : ((m & MEM_MASK) + (s) >= CP_FIFO_END))

#define SCREEN_WIDTH		640
#define SCREEN_HEIGHT		480


__s32 (*gpop[0x100]) (__u32 mem);

GXState gxs;


// pushed matrix indices
int pn_matrix_index;
int tex_matrix_index[8];
// dequantization factor
float vdq, tdq[8];

int last_gamma;

int wireframe = FALSE;
int tex_reload = FALSE;
int dump_textures = FALSE;
int force_linear = FALSE;
int tex_transparent = FALSE;
int dont_draw = FALSE;
int no_logic_ops = FALSE;
int sw_fullbright = FALSE;

TextureTag textags[0x1000];
unsigned int ntextags = 0;
TextureTag *texactive[8] = {0};

int texenabled[8];

// for texture conversion
__u8 texbuff[1024*1024*4];

float real_height = 480, real_width = 640;


inline __u32 color_unpack_rgb565 (__u32 X)
{
#ifdef LIL_ENDIAN
	return (((X & 0xf800) >>  8)  | ((X & 0xe000) >> 13)  |
				  ((X & 0x07e0) <<  5)  | ((X & 0x0600) >>  1)  |
					((X & 0x001f) << 19)  | ((X & 0x001c) << 14)) |
					MASK_ALPHA;
#else
	return (((X & 0xf800) << 16)  | ((X & 0xe000) << 11)  |
				  ((X & 0x07e0) << 13)  | ((X & 0x0600) <<  7)  |
					((X & 0x001f) << 11)  | ((X & 0x001c) <<  6)) |
					MASK_ALPHA;
#endif
}


inline __u32 color_unpack_rgba4 (__u32 X)
{
	__u32 _X = X;


#ifdef LIL_ENDIAN
	_X = ((X & 0x0f00) <<  0) | ((X & 0x00f0) << 12) | ((X & 0x000f) << 24) | ((X & 0xf000) >> 12);
#else
	_X = ((X & 0x0f00) <<  8) | ((X & 0x00f0) <<  4) | ((X & 0x000f) <<  0) | ((X & 0xf000) << 12);
#endif
	return (_X | (_X << 4));
}


inline __u32 color_unpack_rgba6 (__u32 X)
{
#ifdef LIL_ENDIAN
	return ((X & 0xfc0000) >> 16) | ((X & 0xc00000) >> 22) |
				 ((X & 0x03f000) >>  2) | ((X & 0x030000) >>  8) |
				 ((X & 0x000fc0) << 12) | ((X & 0x000c00) <<  6) |
				 ((X & 0x00003f) << 26) | ((X & 0x000030) << 20);
#else
	return ((X & 0xfc0000) <<  8) | ((X & 0xc00000) <<  2) |
				 ((X & 0x03f000) <<  6) | ((X & 0x030000) <<  0) |
				 ((X & 0x000fc0) <<  4) | ((X & 0x000c00) >>  2) |
				 ((X & 0x00003f) <<  2) | ((X & 0x000030) >>  4);
#endif
}

// RGB5 1rrr rrgg gggb bbbb
inline __u32 color_unpack_rgb555 (__u32 X)
{
#ifdef LIL_ENDIAN
	return ((X & 0x7c00) >>  7) | ((X & 0x7000) >> 12) |
				 ((X & 0x03e0) <<  6) | ((X & 0x0380) <<  1) |
				 ((X & 0x001f) << 19) | ((X & 0x001c) << 14);
#else
	return ((X & 0x7c00) << 17) | ((X & 0x7000) << 12) |
				 ((X & 0x03e0) << 14) | ((X & 0x0380) <<  9) |
				 ((X & 0x001f) << 11) | ((X & 0x001c) <<  6);
#endif
}


// RGB4A3  0aaa rrrr gggg bbbb
inline __u32 color_unpack_rgb4a3 (__u32 X)
{
#ifdef LIL_ENDIAN
	return ((X & 0x0f00) >>  4) | ((X & 0x0f00) >>  8) |
				 ((X & 0x00f0) <<  8) | ((X & 0x00f0) <<  4) |
				 ((X & 0x000f) << 20) | ((X & 0x000f) << 16) |
				 ((X & 0x7000) << 17) | ((X & 0x7000) << 13);
#else
	return ((X & 0x0f00) << 20) | ((X & 0x0f00) << 16) |
				 ((X & 0x00f0) << 16) | ((X & 0x00f0) << 12) |
				 ((X & 0x000f) << 12) | ((X & 0x000f) <<  8) |
				 ((X & 0x7000) >>  7) | ((X & 0x7000) >>  8);
#endif
}


inline __u32 color_unpack_rgb5a3 (__u32 X)
{
	if (X & 0x8000)
		return (color_unpack_rgb555 (X) | MASK_ALPHA);
	else
		return color_unpack_rgb4a3 (X);
}


inline __u32 color_unpack_i4 (__u32 X)
{
	X |= X << 4;
	X |= X << 8;
	return (X | (X << 16));
}


inline __u32 color_unpack_i8 (__u32 X)
{
	X |= X << 8;
	return (X | (X << 16));
}


inline __u32 color_unpack_ia4 (__u32 X)
{
	__u32 a = X & 0xf0;


	X &= 0x0f;
	X |= X << 4;
	X |= X << 8;
#ifdef LIL_ENDIAN
	return (X | (X << 8) | (a << 24) | (a << 20));
#else
	return ((X << 16) | (X << 8) | (a >> 4) | a);
#endif
}


inline __u32 color_unpack_ia8 (__u32 X)
{
	__u32 a = X & 0xff00;


	X &= 0x00ff;

#ifdef LIL_ENDIAN
	return (X | (X << 8) | (X << 16) | (a << 16));
#else
	return ((X << 8) | (X << 16) | (X << 24) | (a >> 8));
#endif
}


int gx_parse_list (__u32 start, __u32 length)
{
	__u32 end = start + length;
	int n;


	DEBUG (EVENT_LOG_GX_IMM, "..gx: parsing list (start = %.8x, length = %.8x)", start | 0x80000000, length);

	while (start < end)
	{
		n = GPOP (MEM (start)) (start);
		if (n < 0)
		{
			n = -n;
			if (!((n == 9) && (start + n < end)))
				return -(start - (end - length));
			// otherwise display list problem, continue
		}

		start += n;
	}

	return (start - (end - length));
}


// send position / normal matrix index
unsigned int gx_send_pmidx (__u32 mem)
{
	pn_matrix_index = MEM (mem);
	
	return 1;
}


// send texture matrix index
unsigned int gx_send_t0midx (__u32 mem)
{
	tex_matrix_index[0] = MEM (mem);
	
	return 1;
}


unsigned int gx_send_t1midx (__u32 mem)
{
	tex_matrix_index[1] = MEM (mem);
	
	return 1;
}


unsigned int gx_send_t2midx (__u32 mem)
{
	tex_matrix_index[2] = MEM (mem);
	
	return 1;
}


unsigned int gx_send_t3midx (__u32 mem)
{
	tex_matrix_index[3] = MEM (mem);
	
	return 1;
}


unsigned int gx_send_t4midx (__u32 mem)
{
	tex_matrix_index[4] = MEM (mem);
	
	return 1;
}


unsigned int gx_send_t5midx (__u32 mem)
{
	tex_matrix_index[5] = MEM (mem);
	
	return 1;
}


unsigned int gx_send_t6midx (__u32 mem)
{
	tex_matrix_index[6] = MEM (mem);
	
	return 1;
}


unsigned int gx_send_t7midx (__u32 mem)
{
	tex_matrix_index[7] = MEM (mem);
	
	return 1;
}


// direct position
unsigned int gx_send_pu8_xy (__u32 mem)
{
	__u8 *p = MEM_ADDRESS (mem);


	DEBUG (EVENT_LOG_GX, "....  pos   (%d, %d)", p[0], p[1]);

	glVertex2s (p[0], p[1]);

	return 2;
}


unsigned int gx_send_pu8_xyz (__u32 mem)
{
	__u8 *p = MEM_ADDRESS (mem);


	DEBUG (EVENT_LOG_GX, "....  pos   (%d, %d, %d)", p[0], p[1], p[2]);

	glVertex3s (p[0], p[1], p[2]);

	return 3;
}


unsigned int gx_send_ps8_xy (__u32 mem)
{
	__s8 *p = MEM_ADDRESS (mem);


	DEBUG (EVENT_LOG_GX, "....  pos   (%d, %d)", p[0], p[1]);

	glVertex2s (p[0], p[1]);

	return 2;
}


unsigned int gx_send_ps8_xyz (__u32 mem)
{
	__s8 *p = MEM_ADDRESS (mem);


	DEBUG (EVENT_LOG_GX, "....  pos   (%d, %d, %d)", p[0], p[1], p[2]);

	glVertex3s (p[0], p[1], p[2]);

	return 3;
}


unsigned int gx_send_pu16_xy (__u32 mem)
{
	DEBUG (EVENT_LOG_GX, "....  pos   (%d, %d)", MEMR16 (mem), MEMR16 (mem + 2));

	glVertex2i (MEMR16 (mem), MEMR16 (mem + 2));

	return 4;
}


unsigned int gx_send_pu16_xyz (__u32 mem)
{
	DEBUG (EVENT_LOG_GX, "....  pos   (%d, %d, %d)", MEMR16 (mem), MEMR16 (mem + 2), MEMR16 (mem + 4));

	glVertex3i (MEMR16 (mem), MEMR16 (mem + 2), MEMR16 (mem + 4));

	return 6;
}


unsigned int gx_send_ps16_xy (__u32 mem)
{
	DEBUG (EVENT_LOG_GX, "....  pos   (%d, %d)", MEMSR16 (mem), MEMSR16 (mem + 2));

	glVertex2s (MEMSR16 (mem), MEMSR16 (mem + 2));

	return 4;
}


unsigned int gx_send_ps16_xyz (__u32 mem)
{
	DEBUG (EVENT_LOG_GX, "....  pos   (%d, %d, %d)", MEMSR16 (mem), MEMSR16 (mem + 2), MEMSR16 (mem + 4));

	glVertex3s (MEMSR16 (mem), MEMSR16 (mem + 2), MEMSR16 (mem + 4));

	return 6;
}


unsigned int gx_send_pf_xy (__u32 mem)
{
	DEBUG (EVENT_LOG_GX, "....  pos   (%.2f, %.2f)", MEMRF (mem), MEMRF (mem + 4));

	glVertex2f (MEMRF (mem), MEMRF (mem + 4));

	return 8;
}


unsigned int gx_send_pf_xyz (__u32 mem)
{
	DEBUG (EVENT_LOG_GX, "....  pos   (%.2f, %.2f, %.2f)", MEMRF (mem), MEMRF (mem + 4), MEMRF (mem + 8));

	glVertex3f (MEMRF (mem), MEMRF (mem + 4), MEMRF (mem + 8));

	return 12;
}


// 8 bit indexed position
unsigned int gx_send_pi8u8_xy (__u32 mem)
{
	__u8 *p = MEM_ADDRESS (CP_VERTEX_ARRAY_BASE + MEM (mem) * CP_VERTEX_ARRAY_STRIDE);


	DEBUG (EVENT_LOG_GX, "....  pos   (%d, %d)", p[0], p[1]);

	glVertex2s (p[0], p[1]);

	return 1;
}


unsigned int gx_send_pi8u8_xyz (__u32 mem)
{
	__u8 *p = MEM_ADDRESS (CP_VERTEX_ARRAY_BASE + MEM (mem) * CP_VERTEX_ARRAY_STRIDE);


	DEBUG (EVENT_LOG_GX, "....  pos   (%d, %d, %d)", p[0], p[1], p[2]);

	glVertex3s (p[0], p[1], p[2]);

	return 1;
}


unsigned int gx_send_pi8s8_xy (__u32 mem)
{
	__s8 *p = MEM_ADDRESS (CP_VERTEX_ARRAY_BASE + MEM (mem) * CP_VERTEX_ARRAY_STRIDE);


	DEBUG (EVENT_LOG_GX, "....  pos   (%d, %d)", p[0], p[1]);

	glVertex2s (p[0], p[1]);

	return 1;
}


unsigned int gx_send_pi8s8_xyz (__u32 mem)
{
	__s8 *p = MEM_ADDRESS (CP_VERTEX_ARRAY_BASE + MEM (mem) * CP_VERTEX_ARRAY_STRIDE);


	DEBUG (EVENT_LOG_GX, "....  pos   (%d, %d, %d)", p[0], p[1], p[2]);

	glVertex3s (p[0], p[1], p[2]);

	return 1;
}


unsigned int gx_send_pi8u16_xy (__u32 mem)
{
	__u16 *p = MEM_ADDRESS (CP_VERTEX_ARRAY_BASE + MEM (mem) * CP_VERTEX_ARRAY_STRIDE);


	DEBUG (EVENT_LOG_GX, "....  pos   (%d, %d)", BSWAP16 (p[0]), BSWAP16 (p[1]));

	glVertex2i (BSWAP16 (p[0]), BSWAP16 (p[1]));

	return 1;
}


unsigned int gx_send_pi8u16_xyz (__u32 mem)
{
	__u16 *p = MEM_ADDRESS (CP_VERTEX_ARRAY_BASE + MEM (mem) * CP_VERTEX_ARRAY_STRIDE);


	DEBUG (EVENT_LOG_GX, "....  pos   (%d, %d, %d)", BSWAP16 (p[0]), BSWAP16 (p[1]), BSWAP16 (p[2]));

	glVertex3i (BSWAP16 (p[0]), BSWAP16 (p[1]), BSWAP16 (p[2]));

	return 1;
}


unsigned int gx_send_pi8s16_xy (__u32 mem)
{
	__s16 *p = MEM_ADDRESS (CP_VERTEX_ARRAY_BASE + MEM (mem) * CP_VERTEX_ARRAY_STRIDE);


	DEBUG (EVENT_LOG_GX, "....  pos   (%d, %d)", BSWAPS16 (p[0]), BSWAPS16 (p[1]));

	glVertex2s (BSWAPS16 (p[0]), BSWAPS16 (p[1]));

	return 1;
}


unsigned int gx_send_pi8s16_xyz (__u32 mem)
{
	__s16 *p = MEM_ADDRESS (CP_VERTEX_ARRAY_BASE + MEM (mem) * CP_VERTEX_ARRAY_STRIDE);


	DEBUG (EVENT_LOG_GX, "....  pos   (%d, %d, %d)", BSWAPS16 (p[0]), BSWAPS16 (p[1]), BSWAPS16 (p[2]));

	glVertex3s (BSWAPS16 (p[0]), BSWAPS16 (p[1]), BSWAPS16 (p[2]));

	return 1;
}


unsigned int gx_send_pi8f_xy (__u32 mem)
{
	float *p = MEM_ADDRESS (CP_VERTEX_ARRAY_BASE + MEM (mem) * CP_VERTEX_ARRAY_STRIDE);


	DEBUG (EVENT_LOG_GX, "....  pos   (%.2f, %.2f)", BSWAPF (p[0]), BSWAPF (p[1]));

	glVertex2f (BSWAPF (p[0]), BSWAPF (p[1]));

	return 1;
}


unsigned int gx_send_pi8f_xyz (__u32 mem)
{
	float *p = MEM_ADDRESS (CP_VERTEX_ARRAY_BASE + MEM (mem) * CP_VERTEX_ARRAY_STRIDE);


	DEBUG (EVENT_LOG_GX, "....  pos   (%.2f, %.2f, %.2f)", BSWAPF (p[0]), BSWAPF (p[1]), BSWAPF (p[2]));

	glVertex3f (BSWAPF (p[0]), BSWAPF (p[1]), BSWAPF (p[2]));

	return 1;
}


// 16 bit indexed position
unsigned int gx_send_pi16u8_xy (__u32 mem)
{
	__u8 *p = MEM_ADDRESS (CP_VERTEX_ARRAY_BASE + MEMR16 (mem) * CP_VERTEX_ARRAY_STRIDE);


	DEBUG (EVENT_LOG_GX, "....  pos   (%d, %d)", p[0], p[1]);

	glVertex2s (p[0], p[1]);

	return 2;
}


unsigned int gx_send_pi16u8_xyz (__u32 mem)
{
	__u8 *p = MEM_ADDRESS (CP_VERTEX_ARRAY_BASE + MEMR16 (mem) * CP_VERTEX_ARRAY_STRIDE);


	DEBUG (EVENT_LOG_GX, "....  pos   (%d, %d, %d)", p[0], p[1], p[2]);

	glVertex3s (p[0], p[1], p[2]);

	return 2;
}


unsigned int gx_send_pi16s8_xy (__u32 mem)
{
	__s8 *p = MEM_ADDRESS (CP_VERTEX_ARRAY_BASE + MEMR16 (mem) * CP_VERTEX_ARRAY_STRIDE);


	DEBUG (EVENT_LOG_GX, "....  pos   (%d, %d)", p[0], p[1]);

	glVertex2s (p[0], p[1]);

	return 2;
}


unsigned int gx_send_pi16s8_xyz (__u32 mem)
{
	__s8 *p = MEM_ADDRESS (CP_VERTEX_ARRAY_BASE + MEMR16 (mem) * CP_VERTEX_ARRAY_STRIDE);


	DEBUG (EVENT_LOG_GX, "....  pos   (%d, %d, %d)", p[0], p[1], p[2]);

	glVertex3s (p[0], p[1], p[2]);

	return 2;
}


unsigned int gx_send_pi16u16_xy (__u32 mem)
{
	__u16 *p = MEM_ADDRESS (CP_VERTEX_ARRAY_BASE + MEMR16 (mem) * CP_VERTEX_ARRAY_STRIDE);


	DEBUG (EVENT_LOG_GX, "....  pos   (%d, %d)", BSWAP16 (p[0]), BSWAP16 (p[1]));

	glVertex2i (BSWAP16 (p[0]), BSWAP16 (p[1]));

	return 2;
}


unsigned int gx_send_pi16u16_xyz (__u32 mem)
{
	__u16 *p = MEM_ADDRESS (CP_VERTEX_ARRAY_BASE + MEMR16 (mem) * CP_VERTEX_ARRAY_STRIDE);


	DEBUG (EVENT_LOG_GX, "....  pos   (%d, %d, %d)", BSWAP16 (p[0]), BSWAP16 (p[1]), BSWAP16 (p[2]));

	glVertex3i (BSWAP16 (p[0]), BSWAP16 (p[1]), BSWAP16 (p[2]));

	return 2;
}


unsigned int gx_send_pi16s16_xy (__u32 mem)
{
	__s16 *p = MEM_ADDRESS (CP_VERTEX_ARRAY_BASE + MEMR16 (mem) * CP_VERTEX_ARRAY_STRIDE);


	DEBUG (EVENT_LOG_GX, "....  pos   (%d, %d)", BSWAPS16 (p[0]), BSWAPS16 (p[1]));

	glVertex2s (BSWAPS16 (p[0]), BSWAPS16 (p[1]));

	return 2;
}


unsigned int gx_send_pi16s16_xyz (__u32 mem)
{
	__s16 *p = MEM_ADDRESS (CP_VERTEX_ARRAY_BASE + MEMR16 (mem) * CP_VERTEX_ARRAY_STRIDE);


	DEBUG (EVENT_LOG_GX, "....  pos   (%d, %d, %d)", BSWAPS16 (p[0]), BSWAPS16 (p[1]), BSWAPS16 (p[2]));

	glVertex3s (BSWAPS16 (p[0]), BSWAPS16 (p[1]), BSWAPS16 (p[2]));

	return 2;
}


unsigned int gx_send_pi16f_xy (__u32 mem)
{
	float *p = MEM_ADDRESS (CP_VERTEX_ARRAY_BASE + MEMR16 (mem) * CP_VERTEX_ARRAY_STRIDE);


	DEBUG (EVENT_LOG_GX, "....  pos   (%.2f, %.2f)", BSWAPF (p[0]), BSWAPF (p[1]));

	glVertex2f (BSWAPF (p[0]), BSWAPF (p[1]));

	return 2;
}


unsigned int gx_send_pi16f_xyz (__u32 mem)
{
	float *p = MEM_ADDRESS (CP_VERTEX_ARRAY_BASE + MEMR16 (mem) * CP_VERTEX_ARRAY_STRIDE);


	DEBUG (EVENT_LOG_GX, "....  pos   (%.2f, %.2f, %.2f)", BSWAPF (p[0]), BSWAPF (p[1]), BSWAPF (p[2]));

	glVertex3f (BSWAPF (p[0]), BSWAPF (p[1]), BSWAPF (p[2]));

	return 2;
}


static unsigned int (*gxpos[]) (__u32 mem) =
{
	NULL,

	gx_send_pu8_xy,
	gx_send_ps8_xy,
	gx_send_pu16_xy,
	gx_send_ps16_xy,
	gx_send_pf_xy,
	NULL, NULL, NULL,

	gx_send_pu8_xyz,
	gx_send_ps8_xyz,
	gx_send_pu16_xyz,
	gx_send_ps16_xyz,
	gx_send_pf_xyz,
	NULL, NULL, NULL,

	gx_send_pi8u8_xy,
	gx_send_pi8s8_xy,
	gx_send_pi8u16_xy,
	gx_send_pi8s16_xy,
	gx_send_pi8f_xy,
	NULL, NULL, NULL,

	gx_send_pi8u8_xyz,
	gx_send_pi8s8_xyz,
	gx_send_pi8u16_xyz,
	gx_send_pi8s16_xyz,
	gx_send_pi8f_xyz,
	NULL, NULL, NULL,

	gx_send_pi16u8_xy,
	gx_send_pi16s8_xy,
	gx_send_pi16u16_xy,
	gx_send_pi16s16_xy,
	gx_send_pi16f_xy,
	NULL, NULL, NULL,

	gx_send_pi16u8_xyz,
	gx_send_pi16s8_xyz,
	gx_send_pi16u16_xyz,
	gx_send_pi16s16_xyz,
	gx_send_pi16f_xyz,
	NULL, NULL, NULL,
};


static unsigned int gxpos_size[] =
{
	0,
	2, 2, 4, 4, 8, 0, 0, 0,
	3, 3, 6, 6, 12, 0, 0, 0,
	1, 1, 1, 1, 1, 0, 0, 0,
	1, 1, 1, 1, 1, 0, 0, 0,
	2, 2, 2, 2, 2, 0, 0, 0,
	2, 2, 2, 2, 2, 0, 0, 0,
};


inline void transform_position_xy (float d[2], float v[2], int index)
{
	float *m = &XFMAT (index * 4);


	d[0] = m[0]*v[0] + m[1]*v[1] + m[ 3];
	d[1] = m[4]*v[0] + m[5]*v[1] + m[ 7];
}


inline void transform_position_xyz (float d[3], float v[3], int index)
{
	float *m = &XFMAT (index * 4);


	d[0] = m[0]*v[0] + m[1]*v[1] + m[ 2]*v[2] + m[ 3];
	d[1] = m[4]*v[0] + m[5]*v[1] + m[ 6]*v[2] + m[ 7];
	d[2] = m[8]*v[0] + m[9]*v[1] + m[10]*v[2] + m[11];
}


// direct position transformed by matrix
unsigned int gx_send_pu8_xy_midx (__u32 mem)
{
	float d[2], v[2] =
	{
		vdq * MEM (mem), vdq * MEM (mem + 1),
	};


	transform_position_xy (d, v, pn_matrix_index);
	DEBUG (EVENT_LOG_GX, "....  pos   (%.2f, %.2f)", d[0], d[1]);
	glVertex2fv (d);

	return 2;
}


unsigned int gx_send_pu8_xyz_midx (__u32 mem)
{
	float d[3], v[3] =
	{
		vdq * MEM (mem), vdq * MEM (mem + 1), vdq * MEM (mem + 2),
	};


	transform_position_xyz (d, v, pn_matrix_index);
	DEBUG (EVENT_LOG_GX, "....  pos   (%.2f, %.2f, %.2f)", d[0], d[1], d[2]);
	glVertex3fv (d);

	return 3;
}


unsigned int gx_send_ps8_xy_midx (__u32 mem)
{
	float d[2], v[2] =
	{
		vdq * MEMS (mem), vdq * MEMS (mem + 1),
	};


	transform_position_xy (d, v, pn_matrix_index);
	DEBUG (EVENT_LOG_GX, "....  pos   (%.2f, %.2f)", d[0], d[1]);
	glVertex2fv (d);

	return 2;
}


unsigned int gx_send_ps8_xyz_midx (__u32 mem)
{
	float d[3], v[3] =
	{
		vdq * MEMS (mem), vdq * MEMS (mem + 1), vdq * MEMS (mem + 2),
	};


	transform_position_xyz (d, v, pn_matrix_index);
	DEBUG (EVENT_LOG_GX, "....  pos   (%.2f, %.2f, %.2f)", d[0], d[1], d[2]);
	glVertex3fv (d);

	return 3;
}


unsigned int gx_send_pu16_xy_midx (__u32 mem)
{
	float d[2], v[2] =
	{
		vdq * MEMR16 (mem), vdq * MEMR16 (mem + 2),
	};


	transform_position_xy (d, v, pn_matrix_index);
	DEBUG (EVENT_LOG_GX, "....  pos   (%.2f, %.2f)", d[0], d[1]);
	glVertex2fv (d);

	return 4;
}


unsigned int gx_send_pu16_xyz_midx (__u32 mem)
{
	float d[3], v[3] =
	{
		vdq * MEMR16 (mem), vdq * MEMR16 (mem + 2), vdq * MEMR16 (mem + 4),
	};


	transform_position_xyz (d, v, pn_matrix_index);
	DEBUG (EVENT_LOG_GX, "....  pos   (%.2f, %.2f, %.2f)", d[0], d[1], d[2]);
	glVertex3fv (d);

	return 6;
}


unsigned int gx_send_ps16_xy_midx (__u32 mem)
{
	float d[2], v[2] =
	{
		vdq * MEMSR16 (mem), vdq * MEMSR16 (mem + 2),
	};


	transform_position_xy (d, v, pn_matrix_index);
	DEBUG (EVENT_LOG_GX, "....  pos   (%.2f, %.2f)", d[0], d[1]);
	glVertex2fv (d);

	return 4;
}


unsigned int gx_send_ps16_xyz_midx (__u32 mem)
{
	float d[3], v[3] =
	{
		vdq * MEMSR16 (mem), vdq * MEMSR16 (mem + 2), vdq * MEMSR16 (mem + 4),
	};


	transform_position_xyz (d, v, pn_matrix_index);
	DEBUG (EVENT_LOG_GX, "....  pos   (%.2f, %.2f, %.2f)", d[0], d[1], d[2]);
	glVertex3fv (d);

	return 6;
}


unsigned int gx_send_pf_xy_midx (__u32 mem)
{
	float d[2], v[2] =
	{
		MEMRF (mem), MEMRF (mem + 4),
	};


	transform_position_xy (d, v, pn_matrix_index);
	DEBUG (EVENT_LOG_GX, "....  pos   (%.2f, %.2f)", d[0], d[1]);
	glVertex2fv (d);

	return 8;
}


unsigned int gx_send_pf_xyz_midx (__u32 mem)
{
	float d[3], v[3] =
	{
		MEMRF (mem), MEMRF (mem + 4), MEMRF (mem + 8),
	};


	transform_position_xyz (d, v, pn_matrix_index);
	DEBUG (EVENT_LOG_GX, "....  pos   (%.2f, %.2f, %.2f)", d[0], d[1], d[2]);
	glVertex3fv (d);

	return 12;
}


// 8 bit indexed position
unsigned int gx_send_pi8u8_xy_midx (__u32 mem)
{
	__u8 *p = MEM_ADDRESS (CP_VERTEX_ARRAY_BASE + MEM (mem) * CP_VERTEX_ARRAY_STRIDE);
	float d[2], v[2] =
	{
		vdq * p[0], vdq * p[1],
	};


	transform_position_xy (d, v, pn_matrix_index);
	DEBUG (EVENT_LOG_GX, "....  pos   (%.2f, %.2f)", d[0], d[1]);
	glVertex2fv (d);

	return 1;
}


unsigned int gx_send_pi8u8_xyz_midx (__u32 mem)
{
	__u8 *p = MEM_ADDRESS (CP_VERTEX_ARRAY_BASE + MEM (mem) * CP_VERTEX_ARRAY_STRIDE);
	float d[3], v[3] =
	{
		vdq * p[0], vdq * p[1], vdq * p[2],
	};


	transform_position_xyz (d, v, pn_matrix_index);
	DEBUG (EVENT_LOG_GX, "....  pos   (%.2f, %.2f, %.2f)", d[0], d[1], d[2]);
	glVertex3fv (d);

	return 1;
}


unsigned int gx_send_pi8s8_xy_midx (__u32 mem)
{
	__s8 *p = MEM_ADDRESS (CP_VERTEX_ARRAY_BASE + MEM (mem) * CP_VERTEX_ARRAY_STRIDE);
	float d[2], v[2] =
	{
		vdq * p[0], vdq * p[1],
	};


	transform_position_xy (d, v, pn_matrix_index);
	DEBUG (EVENT_LOG_GX, "....  pos   (%.2f, %.2f)", d[0], d[1]);
	glVertex2fv (d);

	return 1;
}


unsigned int gx_send_pi8s8_xyz_midx (__u32 mem)
{
	__s8 *p = MEM_ADDRESS (CP_VERTEX_ARRAY_BASE + MEM (mem) * CP_VERTEX_ARRAY_STRIDE);
	float d[3], v[3] =
	{
		vdq * p[0], vdq * p[1], vdq * p[2],
	};


	transform_position_xyz (d, v, pn_matrix_index);
	DEBUG (EVENT_LOG_GX, "....  pos   (%.2f, %.2f, %.2f)", d[0], d[1], d[2]);
	glVertex3fv (d);

	return 1;
}


unsigned int gx_send_pi8u16_xy_midx (__u32 mem)
{
	__u16 *p = MEM_ADDRESS (CP_VERTEX_ARRAY_BASE + MEM (mem) * CP_VERTEX_ARRAY_STRIDE);
	float d[2], v[2] =
	{
		vdq * BSWAP16 (p[0]), vdq * BSWAP16 (p[1]),
	};


	transform_position_xy (d, v, pn_matrix_index);
	DEBUG (EVENT_LOG_GX, "....  pos   (%.2f, %.2f)", d[0], d[1]);
	glVertex2fv (d);

	return 1;
}


unsigned int gx_send_pi8u16_xyz_midx (__u32 mem)
{
	__u16 *p = MEM_ADDRESS (CP_VERTEX_ARRAY_BASE + MEM (mem) * CP_VERTEX_ARRAY_STRIDE);
	float d[3], v[3] =
	{
		vdq * BSWAP16 (p[0]), vdq * BSWAP16 (p[1]), vdq * BSWAP16 (p[2]),
	};


	transform_position_xyz (d, v, pn_matrix_index);
	DEBUG (EVENT_LOG_GX, "....  pos   (%.2f, %.2f, %.2f)", d[0], d[1], d[2]);
	glVertex3fv (d);

	return 1;
}


unsigned int gx_send_pi8s16_xy_midx (__u32 mem)
{
	__s16 *p = MEM_ADDRESS (CP_VERTEX_ARRAY_BASE + MEM (mem) * CP_VERTEX_ARRAY_STRIDE);
	float d[2], v[2] =
	{
		vdq * BSWAPS16 (p[0]), vdq * BSWAPS16 (p[1]),
	};


	transform_position_xy (d, v, pn_matrix_index);
	DEBUG (EVENT_LOG_GX, "....  pos   (%.2f, %.2f)", d[0], d[1]);
	glVertex2fv (d);

	return 1;
}


unsigned int gx_send_pi8s16_xyz_midx (__u32 mem)
{
	__s16 *p = MEM_ADDRESS (CP_VERTEX_ARRAY_BASE + MEM (mem) * CP_VERTEX_ARRAY_STRIDE);
	float d[3], v[3] =
	{
		vdq * BSWAPS16 (p[0]), vdq * BSWAPS16 (p[1]), vdq * BSWAPS16 (p[2]),
	};


	transform_position_xyz (d, v, pn_matrix_index);
	DEBUG (EVENT_LOG_GX, "....  pos   (%.2f, %.2f, %.2f)", d[0], d[1], d[2]);
	glVertex3fv (d);

	return 1;
}


unsigned int gx_send_pi8f_xy_midx (__u32 mem)
{
	float *p = MEM_ADDRESS (CP_VERTEX_ARRAY_BASE + MEM (mem) * CP_VERTEX_ARRAY_STRIDE);
	float d[2], v[2] =
	{
		BSWAPF (p[0]), BSWAPF (p[1]),
	};


	transform_position_xy (d, v, pn_matrix_index);
	DEBUG (EVENT_LOG_GX, "....  pos   (%.2f, %.2f)", d[0], d[1]);
	glVertex2fv (d);

	return 1;
}


unsigned int gx_send_pi8f_xyz_midx (__u32 mem)
{
	float *p = MEM_ADDRESS (CP_VERTEX_ARRAY_BASE + MEM (mem) * CP_VERTEX_ARRAY_STRIDE);
	float d[3], v[3] =
	{
		BSWAPF (p[0]), BSWAPF (p[1]), BSWAPF (p[2]),
	};


	transform_position_xyz (d, v, pn_matrix_index);
	DEBUG (EVENT_LOG_GX, "....  pos   (%.2f, %.2f, %.2f)", d[0], d[1], d[2]);
	glVertex3fv (d);

	return 1;
}


// 16 bit indexed position
unsigned int gx_send_pi16u8_xy_midx (__u32 mem)
{
	__u8 *p = MEM_ADDRESS (CP_VERTEX_ARRAY_BASE + MEMR16 (mem) * CP_VERTEX_ARRAY_STRIDE);
	float d[2], v[2] =
	{
		vdq * p[0], vdq * p[1],
	};


	transform_position_xy (d, v, pn_matrix_index);
	DEBUG (EVENT_LOG_GX, "....  pos   (%.2f, %.2f)", d[0], d[1]);
	glVertex2fv (d);

	return 2;
}


unsigned int gx_send_pi16u8_xyz_midx (__u32 mem)
{
	__u8 *p = MEM_ADDRESS (CP_VERTEX_ARRAY_BASE + MEMR16 (mem) * CP_VERTEX_ARRAY_STRIDE);
	float d[3], v[3] =
	{
		vdq * p[0], vdq * p[1], vdq * p[2],
	};


	transform_position_xyz (d, v, pn_matrix_index);
	DEBUG (EVENT_LOG_GX, "....  pos   (%.2f, %.2f, %.2f)", d[0], d[1], d[2]);
	glVertex3fv (d);

	return 2;
}


unsigned int gx_send_pi16s8_xy_midx (__u32 mem)
{
	__s8 *p = MEM_ADDRESS (CP_VERTEX_ARRAY_BASE + MEMR16 (mem) * CP_VERTEX_ARRAY_STRIDE);
	float d[2], v[2] =
	{
		vdq * p[0], vdq * p[1],
	};


	transform_position_xy (d, v, pn_matrix_index);
	DEBUG (EVENT_LOG_GX, "....  pos   (%.2f, %.2f)", d[0], d[1]);
	glVertex2fv (d);

	return 2;
}


unsigned int gx_send_pi16s8_xyz_midx (__u32 mem)
{
	__s8 *p = MEM_ADDRESS (CP_VERTEX_ARRAY_BASE + MEMR16 (mem) * CP_VERTEX_ARRAY_STRIDE);
	float d[3], v[3] =
	{
		vdq * p[0], vdq * p[1], vdq * p[2],
	};


	transform_position_xyz (d, v, pn_matrix_index);
	DEBUG (EVENT_LOG_GX, "....  pos   (%.2f, %.2f, %.2f)", d[0], d[1], d[2]);
	glVertex3fv (d);

	return 2;
}


unsigned int gx_send_pi16u16_xy_midx (__u32 mem)
{
	__u16 *p = MEM_ADDRESS (CP_VERTEX_ARRAY_BASE + MEMR16 (mem) * CP_VERTEX_ARRAY_STRIDE);
	float d[2], v[2] =
	{
		vdq * BSWAP16 (p[0]), vdq * BSWAP16 (p[1]),
	};


	transform_position_xy (d, v, pn_matrix_index);
	DEBUG (EVENT_LOG_GX, "....  pos   (%.2f, %.2f)", d[0], d[1]);
	glVertex2fv (d);

	return 2;
}


unsigned int gx_send_pi16u16_xyz_midx (__u32 mem)
{
	__u16 *p = MEM_ADDRESS (CP_VERTEX_ARRAY_BASE + MEMR16 (mem) * CP_VERTEX_ARRAY_STRIDE);
	float d[3], v[3] =
	{
		vdq * BSWAP16 (p[0]), vdq * BSWAP16 (p[1]), vdq * BSWAP16 (p[2]),
	};


	transform_position_xyz (d, v, pn_matrix_index);
	DEBUG (EVENT_LOG_GX, "....  pos   (%.2f, %.2f, %.2f)", d[0], d[1], d[2]);
	glVertex3fv (d);

	return 2;
}


unsigned int gx_send_pi16s16_xy_midx (__u32 mem)
{
	__s16 *p = MEM_ADDRESS (CP_VERTEX_ARRAY_BASE + MEMR16 (mem) * CP_VERTEX_ARRAY_STRIDE);
	float d[2], v[2] =
	{
		vdq * BSWAPS16 (p[0]), vdq * BSWAPS16 (p[1]),
	};


	transform_position_xy (d, v, pn_matrix_index);
	DEBUG (EVENT_LOG_GX, "....  pos   (%.2f, %.2f)", d[0], d[1]);
	glVertex2fv (d);

	return 2;
}


unsigned int gx_send_pi16s16_xyz_midx (__u32 mem)
{
	__s16 *p = MEM_ADDRESS (CP_VERTEX_ARRAY_BASE + MEMR16 (mem) * CP_VERTEX_ARRAY_STRIDE);
	float d[3], v[3] =
	{
		vdq * BSWAPS16 (p[0]), vdq * BSWAPS16 (p[1]), vdq * BSWAPS16 (p[2]),
	};


	transform_position_xyz (d, v, pn_matrix_index);
	DEBUG (EVENT_LOG_GX, "....  pos   (%.2f, %.2f, %.2f)", d[0], d[1], d[2]);
	glVertex3fv (d);

	return 2;
}


unsigned int gx_send_pi16f_xy_midx (__u32 mem)
{
	float *p = MEM_ADDRESS (CP_VERTEX_ARRAY_BASE + MEMR16 (mem) * CP_VERTEX_ARRAY_STRIDE);
	float d[2], v[2] =
	{
		BSWAPF (p[0]), BSWAPF (p[1]),
	};


	transform_position_xy (d, v, pn_matrix_index);
	DEBUG (EVENT_LOG_GX, "....  pos   (%.2f, %.2f)", d[0], d[1]);
	glVertex2fv (d);

	return 2;
}


unsigned int gx_send_pi16f_xyz_midx (__u32 mem)
{
	float *p = MEM_ADDRESS (CP_VERTEX_ARRAY_BASE + MEMR16 (mem) * CP_VERTEX_ARRAY_STRIDE);
	float d[3], v[3] =
	{
		BSWAPF (p[0]), BSWAPF (p[1]), BSWAPF (p[2]),
	};


	transform_position_xyz (d, v, pn_matrix_index);
	DEBUG (EVENT_LOG_GX, "....  pos   (%.2f, %.2f, %.2f)", d[0], d[1], d[2]);
	glVertex3fv (d);

	return 2;
}


static unsigned int (*gxpos_midx[]) (__u32 mem) =
{
	NULL,

	gx_send_pu8_xy_midx,
	gx_send_ps8_xy_midx,
	gx_send_pu16_xy_midx,
	gx_send_ps16_xy_midx,
	gx_send_pf_xy_midx,
	NULL, NULL, NULL,

	gx_send_pu8_xyz_midx,
	gx_send_ps8_xyz_midx,
	gx_send_pu16_xyz_midx,
	gx_send_ps16_xyz_midx,
	gx_send_pf_xyz_midx,
	NULL, NULL, NULL,

	gx_send_pi8u8_xy_midx,
	gx_send_pi8s8_xy_midx,
	gx_send_pi8u16_xy_midx,
	gx_send_pi8s16_xy_midx,
	gx_send_pi8f_xy_midx,
	NULL, NULL, NULL,

	gx_send_pi8u8_xyz_midx,
	gx_send_pi8s8_xyz_midx,
	gx_send_pi8u16_xyz_midx,
	gx_send_pi8s16_xyz_midx,
	gx_send_pi8f_xyz_midx,
	NULL, NULL, NULL,

	gx_send_pi16u8_xy_midx,
	gx_send_pi16s8_xy_midx,
	gx_send_pi16u16_xy_midx,
	gx_send_pi16s16_xy_midx,
	gx_send_pi16f_xy_midx,
	NULL, NULL, NULL,

	gx_send_pi16u8_xyz_midx,
	gx_send_pi16s8_xyz_midx,
	gx_send_pi16u16_xyz_midx,
	gx_send_pi16s16_xyz_midx,
	gx_send_pi16f_xyz_midx,
	NULL, NULL, NULL,
};


// direct normal
unsigned int gx_send_nu8_xyz (__u32 mem)
{
	return 3;
}


unsigned int gx_send_ns8_xyz (__u32 mem)
{
	return 3;
}


unsigned int gx_send_nu16_xyz (__u32 mem)
{
	return 6;
}


unsigned int gx_send_ns16_xyz (__u32 mem)
{
	return 6;
}


unsigned int gx_send_nf_xyz (__u32 mem)
{
	return 12;
}


unsigned int gx_send_nu8_nbt (__u32 mem)
{
	return 9;
}


unsigned int gx_send_ns8_nbt (__u32 mem)
{
	return 9;
}


unsigned int gx_send_nu16_nbt (__u32 mem)
{
	return 18;
}


unsigned int gx_send_ns16_nbt (__u32 mem)
{
	return 18;
}


unsigned int gx_send_nf_nbt (__u32 mem)
{
	return 36;
}


// 8 bit indexed normal
unsigned int gx_send_ni8u8_xyz (__u32 mem)
{
	return 1;
}


unsigned int gx_send_ni8s8_xyz (__u32 mem)
{
	return 1;
}


unsigned int gx_send_ni8u16_xyz (__u32 mem)
{
	return 1;
}


unsigned int gx_send_ni8s16_xyz (__u32 mem)
{
	return 1;
}


unsigned int gx_send_ni8f_xyz (__u32 mem)
{
	return 1;
}


unsigned int gx_send_ni8u8_nbt (__u32 mem)
{
	return 1;
}


unsigned int gx_send_ni8s8_nbt (__u32 mem)
{
	return 1;
}


unsigned int gx_send_ni8u16_nbt (__u32 mem)
{
	return 1;
}


unsigned int gx_send_ni8s16_nbt (__u32 mem)
{
	return 1;
}


unsigned int gx_send_ni8f_nbt (__u32 mem)
{
	return 1;
}


unsigned int gx_send_ni8u8_nbt3 (__u32 mem)
{
	return 3;
}


unsigned int gx_send_ni8s8_nbt3 (__u32 mem)
{
	return 3;
}


unsigned int gx_send_ni8u16_nbt3 (__u32 mem)
{
	return 3;
}


unsigned int gx_send_ni8s16_nbt3 (__u32 mem)
{
	return 3;
}


unsigned int gx_send_ni8f_nbt3 (__u32 mem)
{
	return 3;
}


// 16 bit indexed normal
unsigned int gx_send_ni16u8_xyz (__u32 mem)
{
	return 2;
}


unsigned int gx_send_ni16s8_xyz (__u32 mem)
{
	return 2;
}


unsigned int gx_send_ni16u16_xyz (__u32 mem)
{
	return 2;
}


unsigned int gx_send_ni16s16_xyz (__u32 mem)
{
	return 2;
}


unsigned int gx_send_ni16f_xyz (__u32 mem)
{
	return 2;
}


unsigned int gx_send_ni16u8_nbt (__u32 mem)
{
	return 2;
}


unsigned int gx_send_ni16s8_nbt (__u32 mem)
{
	return 2;
}


unsigned int gx_send_ni16u16_nbt (__u32 mem)
{
	return 2;
}


unsigned int gx_send_ni16s16_nbt (__u32 mem)
{
	return 2;
}


unsigned int gx_send_ni16f_nbt (__u32 mem)
{
	return 2;
}


unsigned int gx_send_ni16u8_nbt3 (__u32 mem)
{
	return 6;
}


unsigned int gx_send_ni16s8_nbt3 (__u32 mem)
{
	return 6;
}


unsigned int gx_send_ni16u16_nbt3 (__u32 mem)
{
	return 6;
}


unsigned int gx_send_ni16s16_nbt3 (__u32 mem)
{
	return 6;
}


unsigned int gx_send_ni16f_nbt3 (__u32 mem)
{
	return 6;
}


static unsigned int (*gxnormal[]) (__u32 mem) =
{
	NULL,

	gx_send_nu8_xyz,
	gx_send_ns8_xyz,
	gx_send_nu16_xyz,
	gx_send_ns16_xyz,
	gx_send_nf_xyz,
	NULL, NULL, NULL,

	gx_send_nu8_nbt,
	gx_send_ns8_nbt,
	gx_send_nu16_nbt,
	gx_send_ns16_nbt,
	gx_send_nf_nbt,
	NULL, NULL, NULL,

	gx_send_ni8u8_xyz,
	gx_send_ni8s8_xyz,
	gx_send_ni8u16_xyz,
	gx_send_ni8s16_xyz,
	gx_send_ni8f_xyz,
	NULL, NULL, NULL,

	gx_send_ni8u8_nbt,
	gx_send_ni8s8_nbt,
	gx_send_ni8u16_nbt,
	gx_send_ni8s16_nbt,
	gx_send_ni8f_nbt,
	NULL, NULL, NULL,

	gx_send_ni16u8_xyz,
	gx_send_ni16s8_xyz,
	gx_send_ni16u16_xyz,
	gx_send_ni16s16_xyz,
	gx_send_ni16f_xyz,
	NULL, NULL, NULL,

	gx_send_ni16u8_nbt,
	gx_send_ni16s8_nbt,
	gx_send_ni16u16_nbt,
	gx_send_ni16s16_nbt,
	gx_send_ni16f_nbt,
	NULL, NULL, NULL,

	NULL, NULL, NULL, NULL, NULL,
	NULL, NULL, NULL,

	NULL, NULL, NULL, NULL, NULL,
	NULL, NULL, NULL,

	NULL, NULL, NULL, NULL, NULL,
	NULL, NULL, NULL,

	gx_send_ni8u8_nbt3,
	gx_send_ni8s8_nbt3,
	gx_send_ni8u16_nbt3,
	gx_send_ni8s16_nbt3,
	gx_send_ni8f_nbt3,
	NULL, NULL, NULL,

	NULL, NULL, NULL, NULL, NULL,
	NULL, NULL, NULL,

	gx_send_ni16u8_nbt3,
	gx_send_ni16s8_nbt3,
	gx_send_ni16u16_nbt3,
	gx_send_ni16s16_nbt3,
	gx_send_ni16f_nbt3,
	NULL, NULL, NULL,
};


static unsigned int gxnormal_size[] =
{
	0,

	3, 3, 6, 6, 12, 0, 0, 0,
	9, 9, 18, 18, 36, 0, 0, 0,

	1, 1, 1, 1, 1, 0, 0, 0,
	1, 1, 1, 1, 1, 0, 0, 0,

	2, 2, 2, 2, 2, 0, 0, 0,
	2, 2, 2, 2, 2, 0, 0, 0,

	0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0,

	3, 3, 3, 3, 3, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0,
	6, 6, 6, 6, 6, 0, 0, 0,
};


// direct color
unsigned int gx_send_c_rgb565 (__u32 mem)
{
	__u32 c = color_unpack_rgb565 (MEMR16 (mem));


	DEBUG (EVENT_LOG_GX, "....  color (%6.6x)", c);

	glColor3ubv ((__u8 *) &c);

	return 2;
}


unsigned int gx_send_c_rgb8 (__u32 mem)
{
	__u32 c = MEM32 (mem);


	DEBUG (EVENT_LOG_GX, "....  color (%6.6x)", c);

	glColor3ubv ((__u8 *) &c);

	return 3;
}


unsigned int gx_send_c_rgbx8 (__u32 mem)
{
	__u32 c = MEM32 (mem);


	DEBUG (EVENT_LOG_GX, "....  color (%6.6x)", c);

	glColor3ubv ((__u8 *) &c);

	return 4;
}


unsigned int gx_send_c_rgba4 (__u32 mem)
{
	__u32 c = color_unpack_rgba4 (MEMR32 (mem) >> 16);


	DEBUG (EVENT_LOG_GX, "....  color (%8.8x)", c);

	glColor4ubv ((__u8 *) &c);


	return 2;
}


unsigned int gx_send_c_rgba6 (__u32 mem)
{
	__u32 c = color_unpack_rgba6 (MEMR32 (mem) >> 8);


	DEBUG (EVENT_LOG_GX, "....  color (%8.8x)", c);

	glColor4ubv ((__u8 *) &c);

	return 3;
}


unsigned int gx_send_c_rgba8 (__u32 mem)
{
	__u32 c = MEM32 (mem);


	DEBUG (EVENT_LOG_GX, "....  color (%8.8x)", c);

	glColor4ubv ((__u8 *) &c);

	return 4;
}


// 8 bit indexed color
unsigned int gx_send_ci8_rgb565 (__u32 mem)
{
	__u32 c = color_unpack_rgb565 (MEMR16 (CP_COLOR_ARRAY_BASE + MEM (mem) * CP_COLOR_ARRAY_STRIDE));


	DEBUG (EVENT_LOG_GX, "....  color (%6.6x)", c);

	glColor3ubv ((__u8 *) &c);

	return 1;
}


unsigned int gx_send_ci8_rgb8 (__u32 mem)
{
	__u32 c = MEM32 (CP_COLOR_ARRAY_BASE + MEM (mem) * CP_COLOR_ARRAY_STRIDE);


	DEBUG (EVENT_LOG_GX, "....  color (%6.6x)", c);

	glColor3ubv ((__u8 *) &c);

	return 1;
}


unsigned int gx_send_ci8_rgbx8 (__u32 mem)
{
	__u32 c = MEM32 (CP_COLOR_ARRAY_BASE + MEM (mem) * CP_COLOR_ARRAY_STRIDE);


	DEBUG (EVENT_LOG_GX, "....  color (%6.6x)", c);

	glColor3ubv ((__u8 *) &c);

	return 1;
}


unsigned int gx_send_ci8_rgba4 (__u32 mem)
{
	__u32 c = color_unpack_rgba4 (MEMR32 (CP_COLOR_ARRAY_BASE + MEM (mem) * CP_COLOR_ARRAY_STRIDE) >> 16);


	DEBUG (EVENT_LOG_GX, "....  color (%8.8x)", c);

	glColor4ubv ((__u8 *) &c);


	return 1;
}


unsigned int gx_send_ci8_rgba6 (__u32 mem)
{
	__u32 c = color_unpack_rgba6 (MEMR32 (CP_COLOR_ARRAY_BASE + MEM (mem) * CP_COLOR_ARRAY_STRIDE) >> 8);


	DEBUG (EVENT_LOG_GX, "....  color (%8.8x)", c);

	glColor4ubv ((__u8 *) &c);

	return 1;
}


unsigned int gx_send_ci8_rgba8 (__u32 mem)
{
	__u32 c = MEM32 (CP_COLOR_ARRAY_BASE + MEM (mem) * CP_COLOR_ARRAY_STRIDE);


	DEBUG (EVENT_LOG_GX, "....  color (%8.8x)", c);

	glColor4ubv ((__u8 *) &c);

	return 1;
}


// 16 bit indexed color
unsigned int gx_send_ci16_rgb565 (__u32 mem)
{
	__u32 c = color_unpack_rgb565 (MEMR16 (CP_COLOR_ARRAY_BASE + MEMR16 (mem) * CP_COLOR_ARRAY_STRIDE));


	DEBUG (EVENT_LOG_GX, "....  color (%6.6x)", c);

	glColor3ubv ((__u8 *) &c);

	return 2;
}


unsigned int gx_send_ci16_rgb8 (__u32 mem)
{
	__u32 c = MEM32 (CP_COLOR_ARRAY_BASE + MEMR16 (mem) * CP_COLOR_ARRAY_STRIDE);


	DEBUG (EVENT_LOG_GX, "....  color (%6.6x)", c);

	glColor3ubv ((__u8 *) &c);

	return 2;
}


unsigned int gx_send_ci16_rgbx8 (__u32 mem)
{
	__u32 c = MEM32 (CP_COLOR_ARRAY_BASE + MEMR16 (mem) * CP_COLOR_ARRAY_STRIDE);


	DEBUG (EVENT_LOG_GX, "....  color (%6.6x)", c);

	glColor3ubv ((__u8 *) &c);

	return 2;
}


unsigned int gx_send_ci16_rgba4 (__u32 mem)
{
	__u32 c = color_unpack_rgba4 (MEMR32 (CP_COLOR_ARRAY_BASE + MEMR16 (mem) * CP_COLOR_ARRAY_STRIDE) >> 16);


	DEBUG (EVENT_LOG_GX, "....  color (%8.8x)", c);

	glColor4ubv ((__u8 *) &c);


	return 2;
}


unsigned int gx_send_ci16_rgba6 (__u32 mem)
{
	__u32 c = color_unpack_rgba6 (MEMR32 (CP_COLOR_ARRAY_BASE + MEMR16 (mem) * CP_COLOR_ARRAY_STRIDE) >> 8);


	DEBUG (EVENT_LOG_GX, "....  color (%8.8x)", c);

	glColor4ubv ((__u8 *) &c);

	return 2;
}


unsigned int gx_send_ci16_rgba8 (__u32 mem)
{
	__u32 c = MEM32 (CP_COLOR_ARRAY_BASE + MEMR16 (mem) * CP_COLOR_ARRAY_STRIDE);


	DEBUG (EVENT_LOG_GX, "....  color (%8.8x)", c);

	glColor4ubv ((__u8 *) &c);

	return 2;
}


static unsigned int (*gxcol[]) (__u32 mem) =
{
	NULL,

	gx_send_c_rgb565,
	gx_send_c_rgb8,
	gx_send_c_rgbx8,
	gx_send_c_rgba4, gx_send_c_rgba6, gx_send_c_rgba8, NULL, NULL,

	gx_send_c_rgb565, gx_send_c_rgb8, gx_send_c_rgbx8,
	gx_send_c_rgba4,
	gx_send_c_rgba6,
	gx_send_c_rgba8,
	NULL, NULL,

	gx_send_ci8_rgb565,
	gx_send_ci8_rgb8,
	gx_send_ci8_rgbx8,
	gx_send_ci8_rgba4, gx_send_ci8_rgba6, gx_send_ci8_rgba8, NULL, NULL,

	gx_send_ci8_rgb565, gx_send_ci8_rgb8, gx_send_ci8_rgbx8,
	gx_send_ci8_rgba4,
	gx_send_ci8_rgba6,
	gx_send_ci8_rgba8,
	NULL, NULL,

	gx_send_ci16_rgb565,
	gx_send_ci16_rgb8,
	gx_send_ci16_rgbx8,
	gx_send_ci16_rgba4, gx_send_ci16_rgba6, gx_send_ci16_rgba8, NULL, NULL,

	gx_send_ci16_rgb565, gx_send_ci16_rgb8, gx_send_ci16_rgbx8,
	gx_send_ci16_rgba4,
	gx_send_ci16_rgba6,
	gx_send_ci16_rgba8,
	NULL, NULL,
};


static unsigned int gxcol_size[] =
{
	0,

	2, 3, 4, 2, 3, 4, 0, 0,
	2, 3, 4, 2, 3, 4, 0, 0,

	1, 1, 1, 1, 1, 1, 0, 0,
	1, 1, 1, 1, 1, 1, 0, 0,

	2, 2, 2, 2, 2, 2, 0, 0,
	2, 2, 2, 2, 2, 2, 0, 0,
};


// fake color
unsigned int gx_send_fake_c_rgb565 (__u32 mem)
{
	LOG_FAKE (EVENT_LOG_GX, "....  color~(%6.6x)", color_unpack_rgb565 (MEMR16 (mem)));

	return 2;
}


unsigned int gx_send_fake_c_rgb8 (__u32 mem)
{
	LOG_FAKE (EVENT_LOG_GX, "....  color~(%6.6x)", MEM32 (mem));

	return 3;
}


unsigned int gx_send_fake_c_rgbx8 (__u32 mem)
{
	LOG_FAKE (EVENT_LOG_GX, "....  color~(%6.6x)", MEM32 (mem));

	return 4;
}


unsigned int gx_send_fake_c_rgba4 (__u32 mem)
{
	LOG_FAKE (EVENT_LOG_GX, "....  color~(%8.8x)", color_unpack_rgba4 (MEMR32 (mem) >> 16));

	return 2;
}


unsigned int gx_send_fake_c_rgba6 (__u32 mem)
{
	LOG_FAKE (EVENT_LOG_GX, "....  color~(%8.8x)", color_unpack_rgba6 (MEMR32 (mem) >> 8));

	return 3;
}


unsigned int gx_send_fake_c_rgba8 (__u32 mem)
{
	LOG_FAKE (EVENT_LOG_GX, "....  color~(%8.8x)", MEM32 (mem));

	return 4;
}


// 8 bit indexed color
unsigned int gx_send_fake_ci8_rgb565 (__u32 mem)
{
	LOG_FAKE (EVENT_LOG_GX, "....  color~(%6.6x)", color_unpack_rgb565 (MEMR16 (CP_COLOR_ARRAY_BASE + MEM (mem) * CP_COLOR_ARRAY_STRIDE)));

	return 1;
}


unsigned int gx_send_fake_ci8_rgb8 (__u32 mem)
{
	LOG_FAKE (EVENT_LOG_GX, "....  color~(%6.6x)", MEM32 (CP_COLOR_ARRAY_BASE + MEM (mem) * CP_COLOR_ARRAY_STRIDE));

	return 1;
}


unsigned int gx_send_fake_ci8_rgbx8 (__u32 mem)
{
	LOG_FAKE (EVENT_LOG_GX, "....  color~(%6.6x)", MEM32 (CP_COLOR_ARRAY_BASE + MEM (mem) * CP_COLOR_ARRAY_STRIDE));

	return 1;
}


unsigned int gx_send_fake_ci8_rgba4 (__u32 mem)
{
	LOG_FAKE (EVENT_LOG_GX, "....  color~(%8.8x)", color_unpack_rgba4 (MEMR32 (CP_COLOR_ARRAY_BASE + MEM (mem) * CP_COLOR_ARRAY_STRIDE) >> 16));

	return 1;
}


unsigned int gx_send_fake_ci8_rgba6 (__u32 mem)
{
	LOG_FAKE (EVENT_LOG_GX, "....  color~(%8.8x)", color_unpack_rgba6 (MEMR32 (CP_COLOR_ARRAY_BASE + MEM (mem) * CP_COLOR_ARRAY_STRIDE) >> 8));

	return 1;
}


unsigned int gx_send_fake_ci8_rgba8 (__u32 mem)
{
	LOG_FAKE (EVENT_LOG_GX, "....  color~(%8.8x)", MEM32 (CP_COLOR_ARRAY_BASE + MEM (mem) * CP_COLOR_ARRAY_STRIDE));

	return 1;
}


// 16 bit indexed color
unsigned int gx_send_fake_ci16_rgb565 (__u32 mem)
{
	LOG_FAKE (EVENT_LOG_GX, "....  color~(%6.6x)", color_unpack_rgb565 (MEMR16 (CP_COLOR_ARRAY_BASE + MEMR16 (mem) * CP_COLOR_ARRAY_STRIDE)));

	return 2;
}


unsigned int gx_send_fake_ci16_rgb8 (__u32 mem)
{
	LOG_FAKE (EVENT_LOG_GX, "....  color~(%6.6x)", MEM32 (CP_COLOR_ARRAY_BASE + MEMR16 (mem) * CP_COLOR_ARRAY_STRIDE));

	return 2;
}


unsigned int gx_send_fake_ci16_rgbx8 (__u32 mem)
{
	LOG_FAKE (EVENT_LOG_GX, "....  color~(%6.6x)", MEM32 (CP_COLOR_ARRAY_BASE + MEMR16 (mem) * CP_COLOR_ARRAY_STRIDE));

	return 2;
}


unsigned int gx_send_fake_ci16_rgba4 (__u32 mem)
{
	LOG_FAKE (EVENT_LOG_GX, "....  color~(%8.8x)", color_unpack_rgba4 (MEMR32 (CP_COLOR_ARRAY_BASE + MEMR16 (mem) * CP_COLOR_ARRAY_STRIDE) >> 16));

	return 2;
}


unsigned int gx_send_fake_ci16_rgba6 (__u32 mem)
{
	LOG_FAKE (EVENT_LOG_GX, "....  color~(%8.8x)", color_unpack_rgba6 (MEMR32 (CP_COLOR_ARRAY_BASE + MEMR16 (mem) * CP_COLOR_ARRAY_STRIDE) >> 8));

	return 2;
}


unsigned int gx_send_fake_ci16_rgba8 (__u32 mem)
{
	LOG_FAKE (EVENT_LOG_GX, "....  color~(%8.8x)", MEM32 (CP_COLOR_ARRAY_BASE + MEMR16 (mem) * CP_COLOR_ARRAY_STRIDE));

	return 2;
}



static unsigned int (*gxcol_fake[]) (__u32 mem) =
{
	NULL,

	gx_send_fake_c_rgb565,
	gx_send_fake_c_rgb8,
	gx_send_fake_c_rgbx8,
	gx_send_fake_c_rgba4, gx_send_fake_c_rgba6, gx_send_fake_c_rgba8, NULL, NULL,

	gx_send_fake_c_rgb565, gx_send_fake_c_rgb8, gx_send_fake_c_rgbx8,
	gx_send_fake_c_rgba4,
	gx_send_fake_c_rgba6,
	gx_send_fake_c_rgba8,
	NULL, NULL,

	gx_send_fake_ci8_rgb565,
	gx_send_fake_ci8_rgb8,
	gx_send_fake_ci8_rgbx8,
	gx_send_fake_ci8_rgba4, gx_send_fake_ci8_rgba6, gx_send_fake_ci8_rgba8, NULL, NULL,

	gx_send_fake_ci8_rgb565, gx_send_fake_ci8_rgb8, gx_send_fake_ci8_rgbx8,
	gx_send_fake_ci8_rgba4,
	gx_send_fake_ci8_rgba6,
	gx_send_fake_ci8_rgba8,
	NULL, NULL,

	gx_send_fake_ci16_rgb565,
	gx_send_fake_ci16_rgb8,
	gx_send_fake_ci16_rgbx8,
	gx_send_fake_ci16_rgba4, gx_send_fake_ci16_rgba6, gx_send_fake_ci16_rgba8, NULL, NULL,

	gx_send_fake_ci16_rgb565, gx_send_fake_ci16_rgb8, gx_send_fake_ci16_rgbx8,
	gx_send_fake_ci16_rgba4,
	gx_send_fake_ci16_rgba6,
	gx_send_fake_ci16_rgba8,
	NULL, NULL,
};


// direct texture
unsigned int gx_send_tu8_s (__u32 mem)
{
	DEBUG (EVENT_LOG_GX, "....  tex   (%d)", MEM (mem));

	glTexCoord1s (MEM (mem));

	return 1;
}


unsigned int gx_send_tu8_st (__u32 mem)
{
	DEBUG (EVENT_LOG_GX, "....  tex   (%d, %d)", MEM (mem), MEM (mem + 1));

	glTexCoord2s (MEM (mem), MEM (mem + 1));

	return 2;
}


unsigned int gx_send_ts8_s (__u32 mem)
{
	DEBUG (EVENT_LOG_GX, "....  tex   (%d)", MEMS (mem));

	glTexCoord1s (MEMS (mem));

	return 1;
}


unsigned int gx_send_ts8_st (__u32 mem)
{
	DEBUG (EVENT_LOG_GX, "....  tex   (%d, %d)", MEMS (mem), MEMS (mem + 1));

	glTexCoord2s (MEMS (mem), MEMS (mem + 1));

	return 2;
}


unsigned int gx_send_tu16_s (__u32 mem)
{
	DEBUG (EVENT_LOG_GX, "....  tex   (%d)", MEMR16 (mem));

	glTexCoord1i (MEMR16 (mem));

	return 2;
}


unsigned int gx_send_tu16_st (__u32 mem)
{
	DEBUG (EVENT_LOG_GX, "....  tex   (%d, %d)", MEMR16 (mem), MEMR16 (mem + 2));

	glTexCoord2i (MEMR16 (mem), MEMR16 (mem + 2));

	return 4;
}


unsigned int gx_send_ts16_s (__u32 mem)
{
	DEBUG (EVENT_LOG_GX, "....  tex   (%d)", MEMSR16 (mem));

	glTexCoord1s (MEMSR16 (mem));

	return 2;
}


unsigned int gx_send_ts16_st (__u32 mem)
{
	DEBUG (EVENT_LOG_GX, "....  tex   (%d, %d)", MEMSR16 (mem), MEMSR16 (mem + 2));

	glTexCoord2s (MEMSR16 (mem), MEMSR16 (mem + 2));

	return 4;
}


unsigned int gx_send_tf_s (__u32 mem)
{
	DEBUG (EVENT_LOG_GX, "....  tex   (%.2f)", MEMRF (mem));

	glTexCoord1f (MEMRF (mem));

	return 4;
}


unsigned int gx_send_tf_st (__u32 mem)
{
	DEBUG (EVENT_LOG_GX, "....  tex   (%.2f, %.2f)", MEMRF (mem), MEMRF (mem + 4));

	glTexCoord2f (MEMRF (mem), MEMRF (mem + 4));

	return 8;
}


// 8 bit indexed texture
unsigned int gx_send_ti8u8_s (__u32 mem)
{
	__u8 *p = MEM_ADDRESS (CP_TEXTURE_ARRAY_BASE + MEM (mem) * CP_TEXTURE_ARRAY_STRIDE);


	DEBUG (EVENT_LOG_GX, "....  tex   (%d)", p[0]);

	glTexCoord1s (p[0]);

	return 1;
}


unsigned int gx_send_ti8u8_st (__u32 mem)
{
	__u8 *p = MEM_ADDRESS (CP_TEXTURE_ARRAY_BASE + MEM (mem) * CP_TEXTURE_ARRAY_STRIDE);


	DEBUG (EVENT_LOG_GX, "....  tex   (%d, %d)", p[0], p[1]);

	glTexCoord2s (p[0], p[1]);

	return 1;
}


unsigned int gx_send_ti8s8_s (__u32 mem)
{
	__s8 *p = MEM_ADDRESS (CP_TEXTURE_ARRAY_BASE + MEM (mem) * CP_TEXTURE_ARRAY_STRIDE);


	DEBUG (EVENT_LOG_GX, "....  tex   (%d)", p[0]);

	glTexCoord1s (p[0]);

	return 1;
}


unsigned int gx_send_ti8s8_st (__u32 mem)
{
	__s8 *p = MEM_ADDRESS (CP_TEXTURE_ARRAY_BASE + MEM (mem) * CP_TEXTURE_ARRAY_STRIDE);


	DEBUG (EVENT_LOG_GX, "....  tex   (%d, %d)", p[0], p[1]);

	glTexCoord2s (p[0], p[1]);

	return 1;
}


unsigned int gx_send_ti8u16_s (__u32 mem)
{
	__u16 *p = MEM_ADDRESS (CP_TEXTURE_ARRAY_BASE + MEM (mem) * CP_TEXTURE_ARRAY_STRIDE);


	DEBUG (EVENT_LOG_GX, "....  tex   (%d)", BSWAP16 (p[0]));

	glTexCoord1i (BSWAP16 (p[0]));

	return 1;
}


unsigned int gx_send_ti8u16_st (__u32 mem)
{
	__u16 *p = MEM_ADDRESS (CP_TEXTURE_ARRAY_BASE + MEM (mem) * CP_TEXTURE_ARRAY_STRIDE);


	DEBUG (EVENT_LOG_GX, "....  tex   (%d, %d)", BSWAP16 (p[0]), BSWAP16 (p[1]));

	glTexCoord2i (BSWAP16 (p[0]), BSWAP16 (p[1]));

	return 1;
}


unsigned int gx_send_ti8s16_s (__u32 mem)
{
	__s16 *p = MEM_ADDRESS (CP_TEXTURE_ARRAY_BASE + MEM (mem) * CP_TEXTURE_ARRAY_STRIDE);


	DEBUG (EVENT_LOG_GX, "....  tex   (%d)", BSWAPS16 (p[0]));

	glTexCoord1i (BSWAPS16 (p[0]));

	return 1;
}


unsigned int gx_send_ti8s16_st (__u32 mem)
{
	__s16 *p = MEM_ADDRESS (CP_TEXTURE_ARRAY_BASE + MEM (mem) * CP_TEXTURE_ARRAY_STRIDE);


	DEBUG (EVENT_LOG_GX, "....  tex   (%d, %d)", BSWAPS16 (p[0]), BSWAPS16 (p[1]));

	glTexCoord2i (BSWAPS16 (p[0]), BSWAPS16 (p[1]));

	return 1;
}


unsigned int gx_send_ti8f_s (__u32 mem)
{
	float *p = MEM_ADDRESS (CP_TEXTURE_ARRAY_BASE + MEM (mem) * CP_TEXTURE_ARRAY_STRIDE);


	DEBUG (EVENT_LOG_GX, "....  tex   (%.2f)", BSWAPF (p[0]));

	glTexCoord1f (BSWAPF (p[0]));

	return 1;
}


unsigned int gx_send_ti8f_st (__u32 mem)
{
	float *p = MEM_ADDRESS (CP_TEXTURE_ARRAY_BASE + MEM (mem) * CP_TEXTURE_ARRAY_STRIDE);


	DEBUG (EVENT_LOG_GX, "....  tex   (%.2f, %.2f)", BSWAPF (p[0]), BSWAPF (p[1]));

	glTexCoord2f (BSWAPF (p[0]), BSWAPF (p[1]));

	return 1;
}


// 16 bit indexed texture
unsigned int gx_send_ti16u8_s (__u32 mem)
{
	__u8 *p = MEM_ADDRESS (CP_TEXTURE_ARRAY_BASE + MEMR16 (mem) * CP_TEXTURE_ARRAY_STRIDE);


	DEBUG (EVENT_LOG_GX, "....  tex   (%d)", p[0]);

	glTexCoord1s (p[0]);

	return 2;
}


unsigned int gx_send_ti16u8_st (__u32 mem)
{
	__u8 *p = MEM_ADDRESS (CP_TEXTURE_ARRAY_BASE + MEMR16 (mem) * CP_TEXTURE_ARRAY_STRIDE);


	DEBUG (EVENT_LOG_GX, "....  tex   (%d, %d)", p[0], p[1]);

	glTexCoord2s (p[0], p[1]);

	return 2;
}


unsigned int gx_send_ti16s8_s (__u32 mem)
{
	__s8 *p = MEM_ADDRESS (CP_TEXTURE_ARRAY_BASE + MEMR16 (mem) * CP_TEXTURE_ARRAY_STRIDE);


	DEBUG (EVENT_LOG_GX, "....  tex   (%d)", p[0]);

	glTexCoord1s (p[0]);

	return 2;
}


unsigned int gx_send_ti16s8_st (__u32 mem)
{
	__s8 *p = MEM_ADDRESS (CP_TEXTURE_ARRAY_BASE + MEMR16 (mem) * CP_TEXTURE_ARRAY_STRIDE);


	DEBUG (EVENT_LOG_GX, "....  tex   (%d, %d)", p[0], p[1]);

	glTexCoord2s (p[0], p[1]);

	return 2;
}


unsigned int gx_send_ti16u16_s (__u32 mem)
{
	__u16 *p = MEM_ADDRESS (CP_TEXTURE_ARRAY_BASE + MEMR16 (mem) * CP_TEXTURE_ARRAY_STRIDE);


	DEBUG (EVENT_LOG_GX, "....  tex   (%d)", BSWAP16 (p[0]));

	glTexCoord1i (BSWAP16 (p[0]));

	return 2;
}


unsigned int gx_send_ti16u16_st (__u32 mem)
{
	__u16 *p = MEM_ADDRESS (CP_TEXTURE_ARRAY_BASE + MEMR16 (mem) * CP_TEXTURE_ARRAY_STRIDE);


	DEBUG (EVENT_LOG_GX, "....  tex   (%d, %d)", BSWAP16 (p[0]), BSWAP16 (p[1]));

	glTexCoord2i (BSWAP16 (p[0]), BSWAP16 (p[1]));

	return 2;
}


unsigned int gx_send_ti16s16_s (__u32 mem)
{
	__s16 *p = MEM_ADDRESS (CP_TEXTURE_ARRAY_BASE + MEMR16 (mem) * CP_TEXTURE_ARRAY_STRIDE);


	DEBUG (EVENT_LOG_GX, "....  tex   (%d)", BSWAPS16 (p[0]));

	glTexCoord1i (BSWAPS16 (p[0]));

	return 2;
}


unsigned int gx_send_ti16s16_st (__u32 mem)
{
	__s16 *p = MEM_ADDRESS (CP_TEXTURE_ARRAY_BASE + MEMR16 (mem) * CP_TEXTURE_ARRAY_STRIDE);


	DEBUG (EVENT_LOG_GX, "....  tex   (%d, %d)", BSWAPS16 (p[0]), BSWAPS16 (p[1]));

	glTexCoord2i (BSWAPS16 (p[0]), BSWAPS16 (p[1]));

	return 2;
}


unsigned int gx_send_ti16f_s (__u32 mem)
{
	float *p = MEM_ADDRESS (CP_TEXTURE_ARRAY_BASE + MEMR16 (mem) * CP_TEXTURE_ARRAY_STRIDE);


	DEBUG (EVENT_LOG_GX, "....  tex   (%.2f)", BSWAPF (p[0]));

	glTexCoord1f (BSWAPF (p[0]));

	return 2;
}


unsigned int gx_send_ti16f_st (__u32 mem)
{
	float *p = MEM_ADDRESS (CP_TEXTURE_ARRAY_BASE + MEMR16 (mem) * CP_TEXTURE_ARRAY_STRIDE);


	DEBUG (EVENT_LOG_GX, "....  tex   (%.2f, %.2f)", BSWAPF (p[0]), BSWAPF (p[1]));

	glTexCoord2f (BSWAPF (p[0]), BSWAPF (p[1]));

	return 2;
}


static unsigned int (*gxtex[]) (__u32 mem) =
{
	NULL,

	gx_send_tu8_s,
	gx_send_ts8_s,
	gx_send_tu16_s,
	gx_send_ts16_s,
	gx_send_tf_s,
	NULL, NULL, NULL,

	gx_send_tu8_st,
	gx_send_ts8_st,
	gx_send_tu16_st,
	gx_send_ts16_st,
	gx_send_tf_st,
	NULL, NULL, NULL,

	gx_send_ti8u8_s,
	gx_send_ti8s8_s,
	gx_send_ti8u16_s,
	gx_send_ti8s16_s,
	gx_send_ti8f_s,
	NULL, NULL, NULL,

	gx_send_ti8u8_st,
	gx_send_ti8s8_st,
	gx_send_ti8u16_st,
	gx_send_ti8s16_st,
	gx_send_ti8f_st,
	NULL, NULL, NULL,

	gx_send_ti16u8_s,
	gx_send_ti16s8_s,
	gx_send_ti16u16_s,
	gx_send_ti16s16_s,
	gx_send_ti16f_s,
	NULL, NULL, NULL,

	gx_send_ti16u8_st,
	gx_send_ti16s8_st,
	gx_send_ti16u16_st,
	gx_send_ti16s16_st,
	gx_send_ti16f_st,
	NULL, NULL, NULL,
};


static unsigned int gxtex_size[] =
{
	0,

	1, 1, 2, 2, 4, 0, 0, 0,
	2, 2, 4, 4, 8, 0, 0, 0,

	1, 1, 1, 1, 1, 0, 0, 0,
	1, 1, 1, 1, 1, 0, 0, 0,

	2, 2, 2, 2, 2, 0, 0, 0,
	2, 2, 2, 2, 2, 0, 0, 0,
};


inline void transform_texcoord_s (float d[1], float v[1], int index)
{
	float *m = &XFMAT_TEX (index * 4);


	d[0] = m[0]*v[0] + m[ 3];
}


inline void transform_texcoord_st (float d[2], float v[2], int index)
{
	float *m = &XFMAT_TEX (index * 4);


	d[0] = m[0]*v[0] + m[1]*v[1] + m[ 3];
	d[1] = m[4]*v[0] + m[5]*v[1] + m[ 7];
}


// direct texture transformed by matrix
unsigned int gx_send_tu8_s_midx (__u32 mem)
{
	float d[1], v[1] =
	{
		tdq[0] * MEM (mem),
	};


	transform_texcoord_s (d, v, tex_matrix_index[0]);
	DEBUG (EVENT_LOG_GX, "....  tex   (%.2f)", d[0]);

	glTexCoord1fv (d);

	return 1;
}


unsigned int gx_send_tu8_st_midx (__u32 mem)
{
	float d[2], v[2] =
	{
		tdq[0] * MEM (mem), tdq[0] * MEM (mem + 1),
	};


	transform_texcoord_st (d, v, tex_matrix_index[0]);
	DEBUG (EVENT_LOG_GX, "....  tex   (%.2f, %.2f)", d[0], d[1]);

	glTexCoord2fv (d);

	return 2;
}


unsigned int gx_send_ts8_s_midx (__u32 mem)
{
	float d[1], v[1] =
	{
		tdq[0] * MEMS (mem),
	};


	transform_texcoord_s (d, v, tex_matrix_index[0]);
	DEBUG (EVENT_LOG_GX, "....  tex   (%.2f)", d[0]);

	glTexCoord1fv (d);

	return 1;
}


unsigned int gx_send_ts8_st_midx (__u32 mem)
{
	float d[2], v[2] =
	{
		tdq[0] * MEMS (mem), tdq[0] * MEMS (mem + 1),
	};


	transform_texcoord_st (d, v, tex_matrix_index[0]);
	DEBUG (EVENT_LOG_GX, "....  tex   (%.2f, %.2f)", d[0], d[1]);

	glTexCoord2fv (d);

	return 2;
}


unsigned int gx_send_tu16_s_midx (__u32 mem)
{
	float d[1], v[1] =
	{
		tdq[0] * MEMR16 (mem),
	};


	transform_texcoord_s (d, v, tex_matrix_index[0]);
	DEBUG (EVENT_LOG_GX, "....  tex   (%.2f)", d[0]);

	glTexCoord1fv (d);

	return 2;
}


unsigned int gx_send_tu16_st_midx (__u32 mem)
{
	float d[2], v[2] =
	{
		tdq[0] * MEMR16 (mem), tdq[0] * MEMR16 (mem + 2),
	};


	transform_texcoord_st (d, v, tex_matrix_index[0]);
	DEBUG (EVENT_LOG_GX, "....  tex   (%.2f, %.2f)", d[0], d[1]);

	glTexCoord2fv (d);

	return 4;
}


unsigned int gx_send_ts16_s_midx (__u32 mem)
{
	float d[1], v[1] =
	{
		tdq[0] * MEMSR16 (mem),
	};


	transform_texcoord_s (d, v, tex_matrix_index[0]);
	DEBUG (EVENT_LOG_GX, "....  tex   (%.2f)", d[0]);

	glTexCoord1fv (d);

	return 2;
}


unsigned int gx_send_ts16_st_midx (__u32 mem)
{
	float d[2], v[2] =
	{
		tdq[0] * MEMSR16 (mem), tdq[0] * MEMSR16 (mem + 2),
	};


	transform_texcoord_st (d, v, tex_matrix_index[0]);
	DEBUG (EVENT_LOG_GX, "....  tex   (%.2f, %.2f)", d[0], d[1]);

	glTexCoord2fv (d);

	return 4;
}


unsigned int gx_send_tf_s_midx (__u32 mem)
{
	float d[1], v[1] =
	{
		tdq[0] * MEMRF (mem),
	};


	transform_texcoord_s (d, v, tex_matrix_index[0]);
	DEBUG (EVENT_LOG_GX, "....  tex   (%.2f)", d[0]);

	glTexCoord1fv (d);

	return 4;
}


unsigned int gx_send_tf_st_midx (__u32 mem)
{
	float d[2], v[2] =
	{
		tdq[0] * MEMRF (mem), tdq[0] * MEMRF (mem + 4),
	};


	transform_texcoord_st (d, v, tex_matrix_index[0]);
	DEBUG (EVENT_LOG_GX, "....  tex   (%.2f, %.2f)", d[0], d[1]);

	glTexCoord2fv (d);

	return 8;
}


// 8 bit indexed texture
unsigned int gx_send_ti8u8_s_midx (__u32 mem)
{
	__u8 *p = MEM_ADDRESS (CP_TEXTURE_ARRAY_BASE + MEM (mem) * CP_TEXTURE_ARRAY_STRIDE);
	float d[1], v[1] =
	{
		tdq[0] * p[0],
	};


	transform_texcoord_s (d, v, tex_matrix_index[0]);
	DEBUG (EVENT_LOG_GX, "....  tex   (%.2f)", d[0]);

	glTexCoord1fv (d);

	return 1;
}


unsigned int gx_send_ti8u8_st_midx (__u32 mem)
{
	__u8 *p = MEM_ADDRESS (CP_TEXTURE_ARRAY_BASE + MEM (mem) * CP_TEXTURE_ARRAY_STRIDE);
	float d[2], v[2] =
	{
		tdq[0] * p[0], tdq[0] * p[1],
	};


	transform_texcoord_st (d, v, tex_matrix_index[0]);
	DEBUG (EVENT_LOG_GX, "....  tex   (%.2f, %.2f)", d[0], d[1]);

	glTexCoord2fv (d);

	return 1;
}


unsigned int gx_send_ti8s8_s_midx (__u32 mem)
{
	__s8 *p = MEM_ADDRESS (CP_TEXTURE_ARRAY_BASE + MEM (mem) * CP_TEXTURE_ARRAY_STRIDE);
	float d[1], v[1] =
	{
		tdq[0] * p[0],
	};


	transform_texcoord_s (d, v, tex_matrix_index[0]);
	DEBUG (EVENT_LOG_GX, "....  tex   (%.2f)", d[0]);

	glTexCoord1fv (d);

	return 1;
}


unsigned int gx_send_ti8s8_st_midx (__u32 mem)
{
	__s8 *p = MEM_ADDRESS (CP_TEXTURE_ARRAY_BASE + MEM (mem) * CP_TEXTURE_ARRAY_STRIDE);
	float d[2], v[2] =
	{
		tdq[0] * p[0], tdq[0] * p[1],
	};


	transform_texcoord_st (d, v, tex_matrix_index[0]);
	DEBUG (EVENT_LOG_GX, "....  tex   (%.2f, %.2f)", d[0], d[1]);

	glTexCoord2fv (d);

	return 1;
}


unsigned int gx_send_ti8u16_s_midx (__u32 mem)
{
	__u16 *p = MEM_ADDRESS (CP_TEXTURE_ARRAY_BASE + MEM (mem) * CP_TEXTURE_ARRAY_STRIDE);
	float d[1], v[1] =
	{
		tdq[0] * BSWAP16 (p[0]),
	};


	transform_texcoord_s (d, v, tex_matrix_index[0]);
	DEBUG (EVENT_LOG_GX, "....  tex   (%.2f)", d[0]);

	glTexCoord1fv (d);

	return 1;
}


unsigned int gx_send_ti8u16_st_midx (__u32 mem)
{
	__u16 *p = MEM_ADDRESS (CP_TEXTURE_ARRAY_BASE + MEM (mem) * CP_TEXTURE_ARRAY_STRIDE);
	float d[2], v[2] =
	{
		tdq[0] * BSWAP16 (p[0]), tdq[0] * BSWAP16 (p[1]),
	};


	transform_texcoord_st (d, v, tex_matrix_index[0]);
	DEBUG (EVENT_LOG_GX, "....  tex   (%.2f, %.2f)", d[0], d[1]);

	glTexCoord2fv (d);

	return 1;
}


unsigned int gx_send_ti8s16_s_midx (__u32 mem)
{
	__s16 *p = MEM_ADDRESS (CP_TEXTURE_ARRAY_BASE + MEM (mem) * CP_TEXTURE_ARRAY_STRIDE);
	float d[1], v[1] =
	{
		tdq[0] * BSWAPS16 (p[0]),
	};


	transform_texcoord_s (d, v, tex_matrix_index[0]);
	DEBUG (EVENT_LOG_GX, "....  tex   (%.2f)", d[0]);

	glTexCoord1fv (d);

	return 1;
}


unsigned int gx_send_ti8s16_st_midx (__u32 mem)
{
	__s16 *p = MEM_ADDRESS (CP_TEXTURE_ARRAY_BASE + MEM (mem) * CP_TEXTURE_ARRAY_STRIDE);
	float d[2], v[2] =
	{
		tdq[0] * BSWAPS16 (p[0]), tdq[0] * BSWAPS16 (p[1]),
	};


	transform_texcoord_st (d, v, tex_matrix_index[0]);
	DEBUG (EVENT_LOG_GX, "....  tex   (%.2f, %.2f)", d[0], d[1]);

	glTexCoord2fv (d);

	return 1;
}


unsigned int gx_send_ti8f_s_midx (__u32 mem)
{
	float *p = MEM_ADDRESS (CP_TEXTURE_ARRAY_BASE + MEM (mem) * CP_TEXTURE_ARRAY_STRIDE);
	float d[1], v[1] =
	{
		tdq[0] * BSWAPF (p[0]),
	};


	transform_texcoord_s (d, v, tex_matrix_index[0]);
	DEBUG (EVENT_LOG_GX, "....  tex   (%.2f)", d[0]);

	glTexCoord1fv (d);

	return 1;
}


unsigned int gx_send_ti8f_st_midx (__u32 mem)
{
	float *p = MEM_ADDRESS (CP_TEXTURE_ARRAY_BASE + MEM (mem) * CP_TEXTURE_ARRAY_STRIDE);
	float d[2], v[2] =
	{
		tdq[0] * BSWAPF (p[0]), tdq[0] * BSWAPF (p[1]),
	};


	transform_texcoord_st (d, v, tex_matrix_index[0]);
	DEBUG (EVENT_LOG_GX, "....  tex   (%.2f, %.2f)", d[0], d[1]);

	glTexCoord2fv (d);

	return 1;
}


// 16 bit indexed texture
unsigned int gx_send_ti16u8_s_midx (__u32 mem)
{
	__u8 *p = MEM_ADDRESS (CP_TEXTURE_ARRAY_BASE + MEMR16 (mem) * CP_TEXTURE_ARRAY_STRIDE);
	float d[1], v[1] =
	{
		tdq[0] * p[0],
	};


	transform_texcoord_s (d, v, tex_matrix_index[0]);
	DEBUG (EVENT_LOG_GX, "....  tex   (%.2f)", d[0]);

	glTexCoord1fv (d);

	return 2;
}


unsigned int gx_send_ti16u8_st_midx (__u32 mem)
{
	__u8 *p = MEM_ADDRESS (CP_TEXTURE_ARRAY_BASE + MEMR16 (mem) * CP_TEXTURE_ARRAY_STRIDE);
	float d[2], v[2] =
	{
		tdq[0] * p[0], tdq[0] * p[1],
	};


	transform_texcoord_st (d, v, tex_matrix_index[0]);
	DEBUG (EVENT_LOG_GX, "....  tex   (%.2f, %.2f)", d[0], d[1]);

	glTexCoord2fv (d);

	return 2;
}


unsigned int gx_send_ti16s8_s_midx (__u32 mem)
{
	__s8 *p = MEM_ADDRESS (CP_TEXTURE_ARRAY_BASE + MEMR16 (mem) * CP_TEXTURE_ARRAY_STRIDE);
	float d[1], v[1] =
	{
		tdq[0] * p[0],
	};


	transform_texcoord_s (d, v, tex_matrix_index[0]);
	DEBUG (EVENT_LOG_GX, "....  tex   (%.2f)", d[0]);

	glTexCoord1fv (d);

	return 2;
}


unsigned int gx_send_ti16s8_st_midx (__u32 mem)
{
	__s8 *p = MEM_ADDRESS (CP_TEXTURE_ARRAY_BASE + MEMR16 (mem) * CP_TEXTURE_ARRAY_STRIDE);
	float d[2], v[2] =
	{
		tdq[0] * p[0], tdq[0] * p[1],
	};


	transform_texcoord_st (d, v, tex_matrix_index[0]);
	DEBUG (EVENT_LOG_GX, "....  tex   (%.2f, %.2f)", d[0], d[1]);

	glTexCoord2fv (d);

	return 2;
}


unsigned int gx_send_ti16u16_s_midx (__u32 mem)
{
	__u16 *p = MEM_ADDRESS (CP_TEXTURE_ARRAY_BASE + MEMR16 (mem) * CP_TEXTURE_ARRAY_STRIDE);
	float d[1], v[1] =
	{
		tdq[0] * BSWAP16 (p[0]),
	};


	transform_texcoord_s (d, v, tex_matrix_index[0]);
	DEBUG (EVENT_LOG_GX, "....  tex   (%.2f)", d[0]);

	glTexCoord1fv (d);

	return 2;
}


unsigned int gx_send_ti16u16_st_midx (__u32 mem)
{
	__u16 *p = MEM_ADDRESS (CP_TEXTURE_ARRAY_BASE + MEMR16 (mem) * CP_TEXTURE_ARRAY_STRIDE);
	float d[2], v[2] =
	{
		tdq[0] * BSWAP16 (p[0]), tdq[0] * BSWAP16 (p[1]),
	};


	transform_texcoord_st (d, v, tex_matrix_index[0]);
	DEBUG (EVENT_LOG_GX, "....  tex   (%.2f, %.2f)", d[0], d[1]);

	glTexCoord2fv (d);

	return 2;
}


unsigned int gx_send_ti16s16_s_midx (__u32 mem)
{
	__s16 *p = MEM_ADDRESS (CP_TEXTURE_ARRAY_BASE + MEMR16 (mem) * CP_TEXTURE_ARRAY_STRIDE);
	float d[1], v[1] =
	{
		tdq[0] * BSWAPS16 (p[0]),
	};


	transform_texcoord_s (d, v, tex_matrix_index[0]);
	DEBUG (EVENT_LOG_GX, "....  tex   (%.2f)", d[0]);

	glTexCoord1fv (d);

	return 2;
}


unsigned int gx_send_ti16s16_st_midx (__u32 mem)
{
	__s16 *p = MEM_ADDRESS (CP_TEXTURE_ARRAY_BASE + MEMR16 (mem) * CP_TEXTURE_ARRAY_STRIDE);
	float d[2], v[2] =
	{
		tdq[0] * BSWAPS16 (p[0]), tdq[0] * BSWAPS16 (p[1]),
	};


	transform_texcoord_st (d, v, tex_matrix_index[0]);
	DEBUG (EVENT_LOG_GX, "....  tex   (%.2f, %.2f)", d[0], d[1]);

	glTexCoord2fv (d);

	return 2;
}


unsigned int gx_send_ti16f_s_midx (__u32 mem)
{
	float *p = MEM_ADDRESS (CP_TEXTURE_ARRAY_BASE + MEMR16 (mem) * CP_TEXTURE_ARRAY_STRIDE);
	float d[1], v[1] =
	{
		tdq[0] * BSWAPF (p[0]),
	};


	transform_texcoord_s (d, v, tex_matrix_index[0]);
	DEBUG (EVENT_LOG_GX, "....  tex   (%.2f)", d[0]);

	glTexCoord1fv (d);

	return 2;
}


unsigned int gx_send_ti16f_st_midx (__u32 mem)
{
	float *p = MEM_ADDRESS (CP_TEXTURE_ARRAY_BASE + MEMR16 (mem) * CP_TEXTURE_ARRAY_STRIDE);
	float d[2], v[2] =
	{
		tdq[0] * BSWAPF (p[0]), tdq[0] * BSWAPF (p[1]),
	};


	transform_texcoord_st (d, v, tex_matrix_index[0]);
	DEBUG (EVENT_LOG_GX, "....  tex   (%.2f, %.2f)", d[0], d[1]);

	glTexCoord2fv (d);

	return 2;
}


static unsigned int (*gxtex_midx[]) (__u32 mem) =
{
	NULL,

	gx_send_tu8_s_midx,
	gx_send_ts8_s_midx,
	gx_send_tu16_s_midx,
	gx_send_ts16_s_midx,
	gx_send_tf_s_midx,
	NULL, NULL, NULL,

	gx_send_tu8_st_midx,
	gx_send_ts8_st_midx,
	gx_send_tu16_st_midx,
	gx_send_ts16_st_midx,
	gx_send_tf_st_midx,
	NULL, NULL, NULL,

	gx_send_ti8u8_s_midx,
	gx_send_ti8s8_s_midx,
	gx_send_ti8u16_s_midx,
	gx_send_ti8s16_s_midx,
	gx_send_ti8f_s_midx,
	NULL, NULL, NULL,

	gx_send_ti8u8_st_midx,
	gx_send_ti8s8_st_midx,
	gx_send_ti8u16_st_midx,
	gx_send_ti8s16_st_midx,
	gx_send_ti8f_st_midx,
	NULL, NULL, NULL,

	gx_send_ti16u8_s_midx,
	gx_send_ti16s8_s_midx,
	gx_send_ti16u16_s_midx,
	gx_send_ti16s16_s_midx,
	gx_send_ti16f_s_midx,
	NULL, NULL, NULL,

	gx_send_ti16u8_st_midx,
	gx_send_ti16s8_st_midx,
	gx_send_ti16u16_st_midx,
	gx_send_ti16s16_st_midx,
	gx_send_ti16f_st_midx,
	NULL, NULL, NULL,
};


// fake texture
unsigned int gx_send_fake_tu8_s (__u32 mem)
{
	return 1;
}


unsigned int gx_send_fake_tu8_st (__u32 mem)
{
	return 2;
}


unsigned int gx_send_fake_ts8_s (__u32 mem)
{
	return 1;
}


unsigned int gx_send_fake_ts8_st (__u32 mem)
{
	return 2;
}


unsigned int gx_send_fake_tu16_s (__u32 mem)
{
	return 2;
}


unsigned int gx_send_fake_tu16_st (__u32 mem)
{
	return 4;
}


unsigned int gx_send_fake_ts16_s (__u32 mem)
{
	return 2;
}


unsigned int gx_send_fake_ts16_st (__u32 mem)
{
	return 4;
}


unsigned int gx_send_fake_tf_s (__u32 mem)
{
	return 4;
}


unsigned int gx_send_fake_tf_st (__u32 mem)
{
	return 8;
}


// 8 bit indexed texture
unsigned int gx_send_fake_ti8u8_s (__u32 mem)
{
	return 1;
}


unsigned int gx_send_fake_ti8u8_st (__u32 mem)
{
	return 1;
}


unsigned int gx_send_fake_ti8s8_s (__u32 mem)
{
	return 1;
}


unsigned int gx_send_fake_ti8s8_st (__u32 mem)
{
	return 1;
}


unsigned int gx_send_fake_ti8u16_s (__u32 mem)
{
	return 1;
}


unsigned int gx_send_fake_ti8u16_st (__u32 mem)
{
	return 1;
}


unsigned int gx_send_fake_ti8s16_s (__u32 mem)
{
	return 1;
}


unsigned int gx_send_fake_ti8s16_st (__u32 mem)
{
	return 1;
}


unsigned int gx_send_fake_ti8f_s (__u32 mem)
{
	return 1;
}


unsigned int gx_send_fake_ti8f_st (__u32 mem)
{
	return 1;
}


// 16 bit indexed texture
unsigned int gx_send_fake_ti16u8_s (__u32 mem)
{
	return 2;
}


unsigned int gx_send_fake_ti16u8_st (__u32 mem)
{
	return 2;
}


unsigned int gx_send_fake_ti16s8_s (__u32 mem)
{
	return 2;
}


unsigned int gx_send_fake_ti16s8_st (__u32 mem)
{
	return 2;
}


unsigned int gx_send_fake_ti16u16_s (__u32 mem)
{
	return 2;
}


unsigned int gx_send_fake_ti16u16_st (__u32 mem)
{
	return 2;
}


unsigned int gx_send_fake_ti16s16_s (__u32 mem)
{
	return 2;
}


unsigned int gx_send_fake_ti16s16_st (__u32 mem)
{
	return 2;
}


unsigned int gx_send_fake_ti16f_s (__u32 mem)
{
	return 2;
}


unsigned int gx_send_fake_ti16f_st (__u32 mem)
{
	return 2;
}


static unsigned int (*gxtex_fake[]) (__u32 mem) =
{
	NULL,

	gx_send_fake_tu8_s,
	gx_send_fake_ts8_s,
	gx_send_fake_tu16_s,
	gx_send_fake_ts16_s,
	gx_send_fake_tf_s,
	NULL, NULL, NULL,

	gx_send_fake_tu8_st,
	gx_send_fake_ts8_st,
	gx_send_fake_tu16_st,
	gx_send_fake_ts16_st,
	gx_send_fake_tf_st,
	NULL, NULL, NULL,

	gx_send_fake_ti8u8_s,
	gx_send_fake_ti8s8_s,
	gx_send_fake_ti8u16_s,
	gx_send_fake_ti8s16_s,
	gx_send_fake_ti8f_s,
	NULL, NULL, NULL,

	gx_send_fake_ti8u8_st,
	gx_send_fake_ti8s8_st,
	gx_send_fake_ti8u16_st,
	gx_send_fake_ti8s16_st,
	gx_send_fake_ti8f_st,
	NULL, NULL, NULL,

	gx_send_fake_ti16u8_s,
	gx_send_fake_ti16s8_s,
	gx_send_fake_ti16u16_s,
	gx_send_fake_ti16s16_s,
	gx_send_fake_ti16f_s,
	NULL, NULL, NULL,

	gx_send_fake_ti16u8_st,
	gx_send_fake_ti16s8_st,
	gx_send_fake_ti16u16_st,
	gx_send_fake_ti16s16_st,
	gx_send_fake_ti16f_st,
	NULL, NULL, NULL,
};


void gx_load_matrix (unsigned int index, unsigned int w, unsigned int h)
{
	int i, j;
	float *xm = &XFMAT (index * 4);
	float m[16];
	static float im[16] =
	{
		1, 0, 0, 0,
		0, 1, 0, 0,
		0, 0, 1, 0,
		0, 0, 0, 1,
	};

/*
	// must be transposed
	for (i = 0; i < h; i++)
	{
		for (j = 0; j < w; j++)
			m[j*4 + i] = xm[i*4 + j];

		for (j = w; j < 4; j++)
			m[j*4 + i] = im[i*4 + j];

	}

	for (i = h; i < 4; i++)
		for (j = 0; j < 4; j++)
			m[j*4 + i] = im[i*4 + j];
*/
	// must be transposed
	for (i = 0; i < h; i++)
	{
		for (j = 0; j < w; j++)
			m[j*4 + i] = xm[i*4 + j];

		for (j = w; j < 4; j++)
			m[j*4 + i] = im[i*4 + j];

	}

	for (i = h; i < 4; i++)
		for (j = 0; j < 4; j++)
			m[j*4 + i] = im[i*4 + j];

	glLoadMatrixf (m);
	
	DEBUG (EVENT_LOG_GX, "....  loading matrix (%d, %d):", w, h);
	for (i = 0; i < 4; i++)
		DEBUG (EVENT_LOG_GX, "....  %8.2f   %8.2f   %8.2f   %8.2f", m[i*4 + 0], m[i*4 + 1], m[i*4 + 2], m[i*4 + 3]);
}


TextureTag *texture_tag_fetch (__u32 address, __u32 tlut_address)
{
	int i;


	for (i = 0; i < ntextags; i++)
		if ((textags[i].address == address) &&
				(textags[i].tlut_address == tlut_address))
			return &textags[i];

	return NULL;
}


TextureTag *texture_tag_add (__u32 address, __u32 tlut_address,
														 unsigned int tex, unsigned int type,
														 int width, int height, int format, int tlut_format)
{
	textags[ntextags].address = address;
	textags[ntextags].tlut_address = tlut_address;
	textags[ntextags].tex = tex;
	textags[ntextags].type = type;
	
	textags[ntextags].width = width;
	textags[ntextags].height = height;
	textags[ntextags].format = format;
	textags[ntextags].tlut_format = tlut_format;

	if (type == GL_TEXTURE_2D)
		textags[ntextags].p2 = TRUE;
	else
		textags[ntextags].p2 = FALSE;

	return &textags[ntextags++];
}


void texture_tag_destroy (TextureTag *tag)
{
	tag->address = tag->tlut_address = 0;
	
	glDeleteTextures (1, &tag->tex);
}


#define TEX_I4_TILE_WIDTH				8
#define TEX_I4_TILE_HEIGHT			8
// i4 -> rgba
// 8/8/4
void gx_convert_texture_i4_rgba (__u8 *src, int width, int height, __u16 *tlut, int format, TexFormat *tf)
{
	__u32 *dst = (__u32 *) texbuff;
	int i, j, k, l, ofs, maxl;


	for (i = 0; i < height; i += TEX_I4_TILE_HEIGHT)
	{
		for (j = 0; j < width; j += TEX_I4_TILE_WIDTH)
		{
			for (k = 0, ofs = 0; k < TEX_I4_TILE_HEIGHT; k++)
			{
				maxl = MIN (width - j, TEX_I4_TILE_WIDTH);

				for (l = 0; l < maxl; l += 2)
				{
					dst[ofs + l + 0] = color_unpack_i4 ((*src & 0xf0) >> 4);
					if (l + 1 < maxl)
						dst[ofs + l + 1] = color_unpack_i4 ((*src & 0x0f) >> 0);
					src++;
				}

				ofs += width;
				src += (TEX_I4_TILE_WIDTH - maxl)/2;
			}

			dst += maxl;
		}

		dst += width * (TEX_I4_TILE_HEIGHT - 1);
	}

	tf->format = GL_RGBA;
	tf->type = GL_UNSIGNED_BYTE;
	tf->internal_format = GL_RGBA;
}


// doesn't work
void gx_convert_texture_i4 (__u8 *src, int width, int height, __u16 *tlut, int format, TexFormat *tf)
{
	__u8 *dst = (__u8 *) texbuff;
	int i, j, k, l, w4, ofs, maxl;


	if (is_power_of_two (width))
		w4 = 0;
	else
		w4 = (4 - (width & 3)) & 3;

	for (i = 0; i < height; i += TEX_I4_TILE_HEIGHT)
	{
		for (j = 0; j < width; j += TEX_I4_TILE_WIDTH)
		{
			for (k = 0, ofs = 0; k < TEX_I4_TILE_HEIGHT; k++)
			{
				maxl = MIN (width - j, TEX_I4_TILE_WIDTH);

				for (l = 0; l < maxl; l += 2)
				{
					dst[ofs + l + 0]= (*src & 0xf0);
					dst[ofs + l + 1]= (*src & 0x0f) << 4;
					src++;
				}

				ofs += width + w4;
				src += (TEX_I4_TILE_WIDTH - maxl)/2;
			}

			dst += maxl;
		}

		dst += (width + w4) * (TEX_I4_TILE_HEIGHT - 1) + w4;
	}

	tf->format = GL_INTENSITY;
	tf->type = GL_UNSIGNED_BYTE;
	tf->internal_format = GL_INTENSITY4;
}


#define TEX_I8_TILE_WIDTH				8
#define TEX_I8_TILE_HEIGHT			4
// i8 -> rgba
// 8/4/8
void gx_convert_texture_i8_rgba (__u8 *src, int width, int height, __u16 *tlut, int format, TexFormat *tf)
{
	__u32 *dst = (__u32 *) texbuff;
	int i, j, k, l, ofs, maxl;


	for (i = 0; i < height; i += TEX_I8_TILE_HEIGHT)
	{
		for (j = 0; j < width; j += TEX_I8_TILE_WIDTH)
		{
			for (k = 0, ofs = 0; k < TEX_I8_TILE_HEIGHT; k++)
			{
				maxl = MIN (width - j, TEX_I8_TILE_WIDTH);

				for (l = 0; l < maxl; l++)
				{
					dst[ofs + l] = color_unpack_i8 (*src) | MASK_ALPHA;
					src++;
				}

				ofs += width;
				src += TEX_I8_TILE_WIDTH - maxl;
			}

			dst += maxl;
		}

		dst += width * (TEX_I8_TILE_HEIGHT - 1);
	}

	tf->format = GL_RGBA;
	tf->type = GL_UNSIGNED_BYTE;
	tf->internal_format = GL_RGBA;
}


// doesn't work
void gx_convert_texture_i8 (__u8 *src, int width, int height, __u16 *tlut, int format, TexFormat *tf)
{
	__u8 *dst = (__u8 *) texbuff;
	int i, j, k, w4, ofs, maxl;


	if (is_power_of_two (width))
		w4 = 0;
	else
		w4 = (4 - (width & 3)) & 3;

	for (i = 0; i < height; i += TEX_I8_TILE_HEIGHT)
	{
		for (j = 0; j < width; j += TEX_I8_TILE_WIDTH)
		{
			for (k = 0, ofs = 0; k < TEX_I8_TILE_HEIGHT; k++)
			{
				maxl = MIN (width - j, TEX_I8_TILE_WIDTH);

				memcpy (&dst[ofs], src, maxl);
				src += TEX_I8_TILE_WIDTH;

				ofs += width + w4;
			}

			dst += maxl;
		}

		dst += (width + w4) * (TEX_I8_TILE_HEIGHT - 1) + w4;
	}

	tf->format = GL_INTENSITY;
	tf->type = GL_UNSIGNED_BYTE;
	tf->internal_format = GL_INTENSITY8;
}


#define TEX_IA4_TILE_WIDTH				8
#define TEX_IA4_TILE_HEIGHT				4
// ia4 -> rgba
// 8/4/8
void gx_convert_texture_ia4_rgba (__u8 *src, int width, int height, __u16 *tlut, int format, TexFormat *tf)
{
	__u32 *dst = (__u32 *) texbuff;
	int i, j, k, l, ofs, maxl;


	for (i = 0; i < height; i += TEX_IA4_TILE_HEIGHT)
	{
		for (j = 0; j < width; j += TEX_IA4_TILE_WIDTH)
		{
			for (k = 0, ofs = 0; k < TEX_IA4_TILE_HEIGHT; k++)
			{
				maxl = MIN (width - j, TEX_IA4_TILE_WIDTH);

				for (l = 0; l < maxl; l++)
				{
					dst[ofs + l + 0] = color_unpack_ia4 (*src);
					src++;
				}

				ofs += width;
				src += TEX_IA4_TILE_WIDTH - maxl;
			}

			dst += maxl;
		}

		dst += width * (TEX_IA4_TILE_HEIGHT - 1);
	}

	tf->format = GL_RGBA;
	tf->type = GL_UNSIGNED_BYTE;
	tf->internal_format = GL_RGBA;
}


// GL_LUMINANCE4_ALPHA4 is not working as it should (no GL_UNSIGNED_BYTE_4_4 ?)
void gx_convert_texture_ia4 (__u8 *src, int width, int height, __u16 *tlut, int format, TexFormat *tf)
{
	__u16 *dst = (__u16 *) texbuff;
	int i, j, k, l, w4, ofs, maxl;


	if (is_power_of_two (width))
		w4 = 0;
	else
		w4 = (2 - (width & 1)) & 1;

	for (i = 0; i < height; i += TEX_IA4_TILE_HEIGHT)
	{
		for (j = 0; j < width; j += TEX_IA4_TILE_WIDTH)
		{
			for (k = 0, ofs = 0; k < TEX_IA4_TILE_HEIGHT; k++)
			{
				maxl = MIN (width - j, TEX_IA4_TILE_WIDTH);

				for (l = 0; l < maxl; l++)
				{
					dst[ofs + l] = ((*src & 0x0f) << 4) | ((*src & 0xf0) <<  8);
					src++;
				}

				ofs += width + w4;
				src += TEX_IA4_TILE_WIDTH - maxl;
			}

			dst += maxl;
		}

		dst += (width + w4) * (TEX_IA4_TILE_HEIGHT - 1) + w4;
	}

	tf->format = GL_LUMINANCE_ALPHA;
	tf->type = GL_UNSIGNED_BYTE;
	tf->internal_format = GL_LUMINANCE4_ALPHA4;
}


#define TEX_IA8_TILE_WIDTH				4
#define TEX_IA8_TILE_HEIGHT				4
// ia8 -> rgba
// 4/4/16
void gx_convert_texture_ia8_rgba (__u8 *data, int width, int height, __u16 *tlut, int format, TexFormat *tf)
{
	__u32 *dst = (__u32 *) texbuff;
	__u16 *src = (__u16 *) data;
	int i, j, k, l, ofs, maxl;


	for (i = 0; i < height; i += TEX_IA8_TILE_HEIGHT)
	{
		for (j = 0; j < width; j += TEX_IA8_TILE_WIDTH)
		{
			for (k = 0, ofs = 0; k < TEX_IA8_TILE_HEIGHT; k++)
			{
				maxl = MIN (width - j, TEX_IA8_TILE_WIDTH);

				for (l = 0; l < maxl; l++)
				{
					dst[ofs + l] = color_unpack_ia8 (BSWAP16 (*src));
					src++;
				}

				ofs += width;
				src += TEX_IA8_TILE_WIDTH - maxl;
			}

			dst += maxl;
		}

		dst += width * (TEX_IA8_TILE_HEIGHT - 1);
	}

	tf->format = GL_RGBA;
	tf->type = GL_UNSIGNED_BYTE;
	tf->internal_format = GL_RGBA;
}


void gx_convert_texture_ia8 (__u8 *data, int width, int height, __u16 *tlut, int format, TexFormat *tf)
{
	__u16 *dst = (__u16 *) texbuff;
	__u16 *src = (__u16 *) data;
	int i, j, k, l, w4, ofs, maxl;


	if (is_power_of_two (width))
		w4 = 0;
	else
		w4 = (2 - (width & 1)) & 1;

	for (i = 0; i < height; i += TEX_IA8_TILE_HEIGHT)
	{
		for (j = 0; j < width; j += TEX_IA8_TILE_WIDTH)
		{
			for (k = 0, ofs = 0; k < TEX_IA8_TILE_HEIGHT; k++)
			{
				maxl = MIN (width - j, TEX_IA8_TILE_WIDTH);

				for (l = 0; l < maxl; l++)
				{
					dst[ofs + l] = BSWAP16 (*src);
					src++;
				}

				ofs += width + w4;
				src += TEX_IA8_TILE_WIDTH - maxl;
			}

			dst += maxl;
		}

		dst += (width + w4) * (TEX_IA8_TILE_HEIGHT - 1) + w4;
	}

	tf->format = GL_LUMINANCE_ALPHA;
	tf->type = GL_UNSIGNED_BYTE;
	tf->internal_format = GL_LUMINANCE8_ALPHA8;
}


#define TEX_RGB565_TILE_WIDTH				4
#define TEX_RGB565_TILE_HEIGHT			4
// rgb565 -> rgba
// 4/4/16
void gx_convert_texture_rgb565_rgba (__u8 *data, int width, int height, __u16 *tlut, int format, TexFormat *tf)
{
	__u32 *dst = (__u32 *) texbuff;
	__u16 *src = (__u16 *) data;
	int i, j, k, l, ofs, maxl;


	for (i = 0; i < height; i += TEX_RGB565_TILE_HEIGHT)
	{
		for (j = 0; j < width; j += TEX_RGB565_TILE_WIDTH)
		{
			for (k = 0, ofs = 0; k < TEX_RGB565_TILE_HEIGHT; k++)
			{
				maxl = MIN (width - j, TEX_RGB565_TILE_WIDTH);

				for (l = 0; l < maxl; l++)
				{
					dst[ofs + l] = color_unpack_rgb565 (BSWAP16 (*src));
					src++;
				}

				ofs += width;
				src += TEX_RGB565_TILE_WIDTH - maxl;
			}

			dst += maxl;
		}

		dst += width * (TEX_RGB565_TILE_HEIGHT - 1);
	}

	tf->format = GL_RGBA;
	tf->type = GL_UNSIGNED_BYTE;
	tf->internal_format = GL_RGBA;
}


void gx_convert_texture_rgb565 (__u8 *data, int width, int height, __u16 *tlut, int format, TexFormat *tf)
{
	__u16 *dst = (__u16 *) texbuff;
	__u16 *src = (__u16 *) data;
	int i, j, k, l, w4, ofs, maxl;


	if (is_power_of_two (width))
		w4 = 0;
	else
		w4 = (2 - (width & 1)) & 1;

	for (i = 0; i < height; i += TEX_RGB565_TILE_HEIGHT)
	{
		for (j = 0; j < width; j += TEX_RGB565_TILE_WIDTH)
		{
			for (k = 0, ofs = 0; k < TEX_RGB565_TILE_HEIGHT; k++)
			{
				maxl = MIN (width - j, TEX_RGB565_TILE_WIDTH);

				for (l = 0; l < maxl; l++)
				{
					dst[ofs + l] = BSWAP16 (*src);
					src++;
				}

				ofs += width + w4;
				src += TEX_RGB565_TILE_WIDTH - maxl;
			}

			dst += maxl;
		}

		dst += (width + w4) * (TEX_RGB565_TILE_HEIGHT - 1) + w4;
	}

	tf->format = GL_RGB;
	tf->type = GL_UNSIGNED_SHORT_5_6_5;
	tf->internal_format = GL_RGB8;
}


#define TEX_RGB5A3_TILE_WIDTH				4
#define TEX_RGB5A3_TILE_HEIGHT			4
// rgb5a3 -> rgba
// 4/4/16
void gx_convert_texture_rgb5a3_rgba (__u8 *data, int width, int height, __u16 *tlut, int format, TexFormat *tf)
{
	__u32 *dst = (__u32 *) texbuff;
	__u16 *src = (__u16 *) data;
	int i, j, k, l, ofs, maxl;


	for (i = 0; i < height; i += TEX_RGB5A3_TILE_HEIGHT)
	{
		for (j = 0; j < width; j += TEX_RGB5A3_TILE_WIDTH)
		{
			for (k = 0, ofs = 0; k < TEX_RGB5A3_TILE_HEIGHT; k++)
			{
				maxl = MIN (width - j, TEX_RGB5A3_TILE_WIDTH);

				for (l = 0; l < maxl; l++)
				{
					dst[ofs + l] = color_unpack_rgb5a3 (BSWAP16 (*src));
					src++;
				}

				ofs += width;
				src += TEX_RGB5A3_TILE_WIDTH - maxl;
			}

			dst += maxl;
		}

		dst += width * (TEX_RGB5A3_TILE_HEIGHT - 1);
	}

	tf->format = GL_RGBA;
	tf->type = GL_UNSIGNED_BYTE;
	tf->internal_format = GL_RGBA;
}


#define TEX_RGBA8_TILE_WIDTH			4
#define TEX_RGBA8_TILE_HEIGHT			4
// rgba8 -> rgba
// 4/4/32 (two cache lines)
void gx_convert_texture_rgba8_rgba (__u8 *data, int width, int height, __u16 *tlut, int format, TexFormat *tf)
{
	__u32 *dst = (__u32 *) texbuff;
	__u16 *src = (__u16 *) data;
	int i, j, k, l, ofs, maxl;
	__u32 t;


	for (i = 0; i < height; i += TEX_RGBA8_TILE_HEIGHT)
	{
		for (j = 0; j < width; j += TEX_RGBA8_TILE_WIDTH)
		{
			for (k = 0, ofs = 0; k < TEX_RGBA8_TILE_HEIGHT; k++)
			{
				maxl = MIN (width - j, TEX_RGBA8_TILE_WIDTH);

				for (l = 0; l < maxl; l++)
				{
					t = (BSWAP16 (src[0]) << 16) | BSWAP16 (src[16]);
					dst[ofs + l] = BSWAP32 ((t << 8) | ((t >> 24) & 0xff));
					src++;
				}

				ofs += width;
				src += TEX_RGBA8_TILE_WIDTH - maxl;
			}

			dst += maxl;
			src += 16;
		}

		dst += width * (TEX_RGBA8_TILE_HEIGHT - 1);
	}

	tf->format = GL_RGBA;
	tf->type = GL_UNSIGNED_INT_8_8_8_8_REV;
	tf->internal_format = GL_RGBA8;
}


__u32 (*tlut_unpack[]) (__u32 c) =
{
	color_unpack_ia8,
	color_unpack_rgb565,
	color_unpack_rgb5a3,
};

#define TEX_CI4_TILE_WIDTH				8
#define TEX_CI4_TILE_HEIGHT				8
// ci4 -> rgba
// 8/8/4
void gx_convert_texture_ci4_rgba (__u8 *src, int width, int height, __u16 *tlut, int format, TexFormat *tf)
{
	__u32 *dst = (__u32 *) texbuff;
	int i, j, k, l, ofs, maxl;
	__u32 (*color_unpack) (__u32 c) = tlut_unpack[format & 3];


//	tlut += 0xf0;

	for (i = 0; i < height; i += TEX_CI4_TILE_HEIGHT)
	{
		for (j = 0; j < width; j += TEX_CI4_TILE_WIDTH)
		{
			for (k = 0, ofs = 0; k < TEX_CI4_TILE_HEIGHT; k++)
			{
				maxl = MIN (width - j, TEX_CI4_TILE_WIDTH);

				for (l = 0; l < maxl; l += 2)
				{
					dst[ofs + l + 0] = color_unpack (BSWAP16 (tlut[(*src & 0xf0) >> 4]));
					if (l + 1 < maxl)
						dst[ofs + l + 1] = color_unpack (BSWAP16 (tlut[(*src & 0x0f) >> 0]));
					src++;
				}

				ofs += width;
				src += (TEX_CI4_TILE_WIDTH - maxl)/2;
			}

			dst += maxl;
		}

		dst += width * (TEX_CI4_TILE_HEIGHT - 1);
	}

	tf->format = GL_RGBA;
	tf->type = GL_UNSIGNED_BYTE;
	tf->internal_format = GL_RGBA;
}


#define TEX_CI8_TILE_WIDTH				8
#define TEX_CI8_TILE_HEIGHT				4
// ci8 -> rgba
// 8/4/8
void gx_convert_texture_ci8_rgba (__u8 *src, int width, int height, __u16 *tlut, int format, TexFormat *tf)
{
	__u32 *dst = (__u32 *) texbuff;
	int i, j, k, l, ofs, maxl;
	__u32 (*color_unpack) (__u32 c) = tlut_unpack[format & 3];


	for (i = 0; i < height; i += TEX_CI8_TILE_HEIGHT)
	{
		for (j = 0; j < width; j += TEX_CI8_TILE_WIDTH)
		{
			for (k = 0, ofs = 0; k < TEX_CI8_TILE_HEIGHT; k++)
			{
				maxl = MIN (width - j, TEX_CI8_TILE_WIDTH);

				for (l = 0; l < maxl; l++)
				{
					dst[ofs + l] = color_unpack (BSWAP16 (tlut[*src]));
					src++;
				}

				ofs += width;
				src += TEX_CI8_TILE_WIDTH - maxl;
			}

			dst += maxl;
		}

		dst += width * (TEX_CI8_TILE_HEIGHT - 1);
	}

	tf->format = GL_RGBA;
	tf->type = GL_UNSIGNED_BYTE;
	tf->internal_format = GL_RGBA;
}


#define TEX_CI14X2_TILE_WIDTH				4
#define TEX_CI14X2_TILE_HEIGHT				4
// ci14x2 -> rgba
// 4/4/16
void gx_convert_texture_ci14x2_rgba (__u8 *data, int width, int height, __u16 *tlut, int format, TexFormat *tf)
{
	__u32 *dst = (__u32 *) texbuff;
	__u16 *src = (__u16 *) data;
	int i, j, k, l, ofs, maxl;
	__u32 (*color_unpack) (__u32 c) = tlut_unpack[format & 3];


	for (i = 0; i < height; i += TEX_CI14X2_TILE_HEIGHT)
	{
		for (j = 0; j < width; j += TEX_CI14X2_TILE_WIDTH)
		{
			for (k = 0, ofs = 0; k < TEX_CI14X2_TILE_HEIGHT; k++)
			{
				maxl = MIN (width - j, TEX_CI14X2_TILE_WIDTH);

				for (l = 0; l < maxl; l++)
				{
					dst[ofs + l] = color_unpack (BSWAP16 (tlut[BSWAP16 (*src) & 0x3fff]));
					src++;
				}

				ofs += width;
				src += TEX_CI14X2_TILE_WIDTH - maxl;
			}

			dst += maxl;
		}

		dst += width * (TEX_CI14X2_TILE_HEIGHT - 1);
	}

	tf->format = GL_RGBA;
	tf->type = GL_UNSIGNED_BYTE;
	tf->internal_format = GL_RGBA;
}



__u32 icolor (__u32 a, __u32 b, float fa, float fb, float fc)
{
	__u8 *aa = (__u8 *) &a;
	__u8 *bb = (__u8 *) &b;
	__u8 cc[4];
	int i;


	for (i = 0; i < 4; i++)
		cc[i] = ((__u32) aa[i]*fa + (__u32) bb[i]*fb) / fc;

	return (*(__u32 *) cc);
}


#define TEX_CMP_TILE_WIDTH			8
#define TEX_CMP_TILE_HEIGHT			8
// cmp -> rgba
// 8/8/16 (2x2 blocks of 4x4 texels)
void gx_convert_texture_cmp_rgba (__u8 *data, int width, int height, __u16 *tlut, int format, TexFormat *tf)
{
	__u32 *dst = (__u32 *) texbuff;
	__u16 *src = (__u16 *) data;
	int i, j, k, l, n, ofs, maxw;
	__u32 rgb[4];
	__u8 *cm;


	for (i = 0; i < height; i += TEX_CMP_TILE_HEIGHT)
	{
		for (j = 0; j < width; j += TEX_CMP_TILE_WIDTH)
		{
			maxw = MIN (width - j, TEX_CMP_TILE_WIDTH);

			for (k = 0; k < 2; k++)
			{
				for (l = 0; l < 2; l++)
				{
					rgb[0] = color_unpack_rgb565 (BSWAP16 (src[0]));
					rgb[1] = color_unpack_rgb565 (BSWAP16 (src[1]));

					if (BSWAP16 (src[0]) > BSWAP16 (src[1]))
					{
						rgb[2] = icolor (rgb[0], rgb[1], 2, 1, 3) | MASK_ALPHA;
						rgb[3] = icolor (rgb[1], rgb[0], 2, 1, 3) | MASK_ALPHA;
					}
					else
					{
						rgb[2] = icolor (rgb[0], rgb[1], 0.5, 0.5, 1) | MASK_ALPHA;
						rgb[3] = icolor (rgb[1], rgb[0], 2, 1, 3) &~ MASK_ALPHA;
					}

					// color selection (00, 01, 10, 11)
					cm = (__u8 *) &src[2];
					for (n = 0, ofs = l*4; n < 4; n++)
					{
						// one row (4 texels)
						if (maxw > 0 + l*4)
							dst[ofs + 0] = rgb[(cm[n] & 0xc0) >> 6];
						if (maxw > 1 + l*4)
							dst[ofs + 1] = rgb[(cm[n] & 0x30) >> 4];
						if (maxw > 2 + l*4)
							dst[ofs + 2] = rgb[(cm[n] & 0x0c) >> 2];
						if (maxw > 3 + l*4)
							dst[ofs + 3] = rgb[(cm[n] & 0x03) >> 0];

						ofs += width;
					}

					src += 4;
				}

				dst += width * 4;
			}

			dst += maxw - width * 8;
		}

		dst += width * (TEX_CMP_TILE_HEIGHT - 1);
	}

	tf->format = GL_RGBA;
	tf->type = GL_UNSIGNED_BYTE;
	tf->internal_format = GL_RGBA;
}


void (*texconvert_rgba[]) (__u8 *src, int width, int height, __u16 *tlut, int format, TexFormat *tf) =
{
	gx_convert_texture_i4_rgba,
	gx_convert_texture_i8_rgba,
	gx_convert_texture_ia4_rgba,
	gx_convert_texture_ia8_rgba,
	gx_convert_texture_rgb565_rgba,
	gx_convert_texture_rgb5a3_rgba,
	gx_convert_texture_rgba8_rgba,
	NULL,
	gx_convert_texture_ci4_rgba,
	gx_convert_texture_ci8_rgba,
	gx_convert_texture_ci14x2_rgba,
	NULL,
	NULL,
	NULL,
	gx_convert_texture_cmp_rgba,
	NULL,
	
	NULL,
};


void (*texconvert[]) (__u8 *src, int width, int height, __u16 *tlut, int format, TexFormat *tf) =
{
// sumthings wrong here
//	gx_convert_texture_i4,
//	gx_convert_texture_i8,
	gx_convert_texture_i4_rgba,
	gx_convert_texture_i8_rgba,

	gx_convert_texture_ia4,
	gx_convert_texture_ia8,
	gx_convert_texture_rgb565,
	gx_convert_texture_rgb5a3_rgba,
	gx_convert_texture_rgba8_rgba,
	NULL,

	gx_convert_texture_ci4_rgba,
	gx_convert_texture_ci8_rgba,
	gx_convert_texture_ci14x2_rgba,

	NULL,
	NULL,
	NULL,
	gx_convert_texture_cmp_rgba,
	NULL,
	
	NULL,
};


void gx_enable_texture (unsigned int index, int enable)
{
	switch (enable)
	{
		case TEX_DISABLED:
			if (texenabled[index])
			{
				if (texenabled[index] == TEX_ENABLED_P2)
					glDisable (GL_TEXTURE_2D);
				else
					glDisable (GL_TEXTURE_RECTANGLE);
				
				texenabled[index] = TEX_DISABLED;
			}
			break;

		case GL_TEXTURE_2D:	
		case TEX_ENABLED_P2:
			if (texenabled[index] == TEX_ENABLED)
				glDisable (GL_TEXTURE_RECTANGLE);
			
			glEnable (GL_TEXTURE_2D);
			texenabled[index] = TEX_ENABLED_P2;
			break;

		default:
			if (texenabled[index] == TEX_ENABLED_P2)
				glDisable (GL_TEXTURE_2D);
			
			glEnable (GL_TEXTURE_RECTANGLE);
			texenabled[index] = TEX_ENABLED;
			break;
	}
}


void gx_set_texture_mode0 (int index)
{
	unsigned int wrap[] =
	{
		GL_CLAMP, GL_REPEAT, GL_MIRRORED_REPEAT_IBM,
	};
	unsigned int minfilter[] =
	{
	// needs mipmapping
//		GL_NEAREST, GL_NEAREST_MIPMAP_NEAREST, GL_NEAREST_MIPMAP_LINEAR, 0,
//		GL_LINEAR, GL_LINEAR_MIPMAP_NEAREST, GL_LINEAR_MIPMAP_LINEAR, 
		GL_NEAREST, GL_NEAREST, GL_NEAREST, 0,
		GL_LINEAR, GL_LINEAR, GL_LINEAR, 
	};
	unsigned int magfilter[] =
	{
		GL_NEAREST, GL_LINEAR,
	};


	DEBUG (EVENT_LOG_GX, "....  TEX WRAP S %d T %d FILTER MAG %d MIN %d",
				 TEX_MODE_WRAP_S (index), TEX_MODE_WRAP_T (index),
				 TEX_MODE_MAG_FILTER (index), TEX_MODE_MIN_FILTER (index));

	if (!texactive[index])
		return;

	glTexParameteri (texactive[index]->type,
										GL_TEXTURE_WRAP_S, wrap[TEX_MODE_WRAP_S (index)]);
	glTexParameteri (texactive[index]->type,
										GL_TEXTURE_WRAP_T, wrap[TEX_MODE_WRAP_T (index)]);

	if (force_linear)
	{
		glTexParameteri (texactive[index]->type,
											GL_TEXTURE_MIN_FILTER, GL_LINEAR);
		glTexParameteri (texactive[index]->type,
											GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	}
	else
	{
		glTexParameteri (texactive[index]->type,
											GL_TEXTURE_MIN_FILTER, minfilter[TEX_MODE_MIN_FILTER (index)]);
		glTexParameteri (texactive[index]->type,
											GL_TEXTURE_MAG_FILTER, magfilter[TEX_MODE_MAG_FILTER (index)]);
	}
}


void gx_set_texture_mode1 (int index)
{
	DEBUG (EVENT_LOG_GX, "....  TEX LOD MIN %d MAX %d",
				 TEX_MODE_MIN_LOD (index), TEX_MODE_MAX_LOD (index));
}


char *gx_convert_texture (__u32 address, int width, int height, int format,
													__u32 tlut_address, int tlut_format)
{
	TexFormat tf;


	if (!texconvert_rgba[format])
		return NULL;

	texconvert_rgba[format] (MEM_ADDRESS (address),
													width, height,
													(__s16 *) TMEM_ADDRESS (tlut_address),
													tlut_format,
													&tf);

	return texbuff;
}


void gx_dump_active_texture (int index)
{
	char fname[256];
	TexFormat tf;


	if (texactive[index]->format >= 8 && texactive[index]->format <= 10)
		sprintf (fname, "tex@%.8x_%.2d+%d_%.8x.tga",
							texactive[index]->address | 0x80000000, texactive[index]->format,
							texactive[index]->tlut_format, texactive[index]->tlut_address);
	else
		sprintf (fname, "tex@%.8x_%.2d.tga",
						 texactive[index]->address | 0x80000000, texactive[index]->format);

	texconvert_rgba[texactive[index]->format] (
													MEM_ADDRESS (texactive[index]->address),
													texactive[index]->width, texactive[index]->height,
													(__s16 *) TMEM_ADDRESS (texactive[index]->tlut_address),
													texactive[index]->tlut_format,
													&tf);

	video_dump_texture (fname, texbuff, texactive[index]->width, texactive[index]->height);
}


void gx_load_texture (unsigned int index)
{
	__u32 address = TEX_IMAGE_BASE (index);
	__u32 tlut_address = TLUT_TMEM_BASE (index);
	int format = TEX_FORMAT (index);
	int width = TEX_WIDTH (index);
	int height = TEX_HEIGHT (index);
	int tlut_format = TLUT_FORMAT (index);
	unsigned int tex, type;
	TexFormat tf;
	TextureTag *tag;
	

	tag = texture_tag_fetch (address, tlut_address);
	
	if (tag && tex_reload)
	{
		if ((tag->width != width) || (tag->height != height))
		{
			texture_tag_destroy (tag);
			tag = NULL;
		}
		else
			tag->reload = TRUE;
	}
	
	if (tag)
	{
		gx_enable_texture (index, tag->type);
		if (tag != texactive[index])
		{
			glBindTexture (tag->type, tag->tex);
			texactive[index] = tag;
		}

		if (tag->reload)
		{
			texconvert[format] (MEM_ADDRESS (address),
													width, height,
													(__s16 *) TMEM_ADDRESS (tlut_address),
													tlut_format,
													&tf);

			glTexSubImage2D (tag->type, 0, 0, 0, width, height,
											 tf.format, tf.type, texbuff); 
			tag->reload = FALSE;
		}		

		gx_set_texture_mode0 (index);
		return;
	}

	if (!texconvert[format] || !address)
	{
		texactive[index] = NULL;
		return;
	}

	DEBUG (EVENT_LOG_GX, "....  load tex from %.8x W %d H %d F %d tlut %.8x F %d",
				 address | 0x80000000, width, height, format, tlut_address, tlut_format);

	texconvert[format] (MEM_ADDRESS (address),
											width, height,
											(__s16 *) TMEM_ADDRESS (tlut_address),
											tlut_format,
											&tf);

	if (is_power_of_two (width) && is_power_of_two (height))
		type = GL_TEXTURE_2D;
	else
		type = GL_TEXTURE_RECTANGLE;

	// load texture
	gx_enable_texture (0, type);
	glGenTextures (1, &tex);
	glBindTexture (type, tex);

	// add tag
	texactive[index] = texture_tag_add (address, tlut_address, tex, type,
																			width, height, format, tlut_format);

	gx_set_texture_mode0 (index);

	glTexImage2D (type, 0, tf.internal_format, width, height,
								0, tf.format, tf.type, texbuff);

	if (dump_textures)
		gx_dump_active_texture (index);
}


void gx_reload_tlut (__u32 tlut_address)
{
/*
	int i;


	for (i = 0; i < ntextags; i++)
		if (textags[i].tlut_address == tlut_address)
			textags[i].reload = TRUE;
*/
}


unsigned int gx_draw (__u32 mem, int prim, int n, int vat)
{
	unsigned int (*midx[16]) (__u32 mem);
	unsigned int (*vdata[16]) (__u32 mem);
	unsigned int (*pos) (__u32 mem);
	int nm = 0, nv = 0, psize, i, j;
	__u32 start = mem, mp = mem, list_size = 0, color;


	if (!n)
		return 0;

	// matrix indices
	if (VCD_MIDX)
	{
		if (VCD_PMIDX)
			midx[nm++] = gx_send_pmidx;

		if (VCD_T0MIDX)
			midx[nm++] = gx_send_t0midx;

		if (VCD_T1MIDX)
			midx[nm++] = gx_send_t1midx;
		if (VCD_T2MIDX)
			midx[nm++] = gx_send_t2midx;
		if (VCD_T3MIDX)
			midx[nm++] = gx_send_t3midx;
		if (VCD_T4MIDX)
			midx[nm++] = gx_send_t4midx;
		if (VCD_T5MIDX)
			midx[nm++] = gx_send_t5midx;
		if (VCD_T6MIDX)
			midx[nm++] = gx_send_t6midx;
		if (VCD_T7MIDX)
			midx[nm++] = gx_send_t7midx;
		
	}

	list_size += nm;
	list_size += gxnormal_size[GXIDX (NORMAL, vat)];
	list_size += gxcol_size[GXIDX (COL0, vat)];
	list_size += gxcol_size[GXIDX (COL1, vat)];
	list_size += gxtex_size[GXIDX (TEX0, vat)];
	list_size += gxtex_size[GXIDX (TEX1, vat)];
	list_size += gxtex_size[GXIDX (TEX2, vat)];
	list_size += gxtex_size[GXIDX (TEX3, vat)];
	list_size += gxtex_size[GXIDX (TEX4, vat)];
	list_size += gxtex_size[GXIDX (TEX5, vat)];
	list_size += gxtex_size[GXIDX (TEX6, vat)];
	list_size += gxtex_size[GXIDX (TEX7, vat)];
	list_size += gxpos_size[GXIDX (POS, vat)];
	if (EXCEEDES_LIST_BOUNDARY (start, list_size * n + 3))
	// draw commands take 3 bytes and return 3 + gx_draw
		return -(list_size * n + 6);

	DEBUG (EVENT_LOG_GX, "....  VCD MIDX %2.2x POS %d NORMAL %d COL %d|%d TEX %d|%d|%d|%d|%d|%d|%d|%d",
											 VCD_MIDX, VCD_POS, VCD_NORMAL, VCD_COL0, VCD_COL1,
											 VCD_TEX0, VCD_TEX1, VCD_TEX2, VCD_TEX3, VCD_TEX4, VCD_TEX5, VCD_TEX6, VCD_TEX7);

	DEBUG (EVENT_LOG_GX, "....  VAT DQ %d POS %d|%d|%d NORMAL %d|%d COL0 %d|%d COL1 %d|%d",
											 VAT_BYTE_DEQUANT (vat),
											 VAT_POS_CNT (vat), VAT_POS_FMT (vat), VAT_POS_SHFT (vat),
											 VAT_NORMAL_CNT (vat), VAT_NORMAL_FMT (vat),
											 VAT_COL0_CNT (vat), VAT_COL0_FMT (vat), VAT_COL1_CNT (vat), VAT_COL1_FMT (vat));

	DEBUG (EVENT_LOG_GX, "....      TEX 0 %d|%d|%d 1 %d|%d|%d 2 %d|%d|%d 3 %d|%d|%d 4 %d|%d|%d 5 %d|%d|%d 6 %d|%d|%d 7 %d|%d|%d",
											 VAT_TEX0_CNT (vat), VAT_TEX0_FMT (vat), VAT_TEX0_SHFT (vat),
											 VAT_TEX1_CNT (vat), VAT_TEX1_FMT (vat), VAT_TEX1_SHFT (vat),
											 VAT_TEX2_CNT (vat), VAT_TEX2_FMT (vat), VAT_TEX2_SHFT (vat),
											 VAT_TEX3_CNT (vat), VAT_TEX3_FMT (vat), VAT_TEX3_SHFT (vat),
											 VAT_TEX4_CNT (vat), VAT_TEX4_FMT (vat), VAT_TEX4_SHFT (vat),
											 VAT_TEX5_CNT (vat), VAT_TEX5_FMT (vat), VAT_TEX5_SHFT (vat),
											 VAT_TEX6_CNT (vat), VAT_TEX6_FMT (vat), VAT_TEX6_SHFT (vat),
											 VAT_TEX7_CNT (vat), VAT_TEX7_FMT (vat), VAT_TEX7_SHFT (vat));

	DEBUG (EVENT_LOG_GX, "....  ARRAY POS %.8x|%d COL %.8x|%d TEX %.8x|%d",
											 CP_VERTEX_ARRAY_BASE | 0x80000000, CP_VERTEX_ARRAY_STRIDE,
											 CP_COLOR_ARRAY_BASE | 0x80000000, CP_COLOR_ARRAY_STRIDE,
											 CP_TEXTURE_ARRAY_BASE | 0x80000000, CP_TEXTURE_ARRAY_STRIDE);

	DEBUG (EVENT_LOG_GX, "....  HOST COLORS %d NORMALS %d TEXTURES %d",
											 HOST_COLORS, HOST_NORMALS, HOST_TEXTURES);

	if (dont_draw)
	{
		gcube_perf_vertices (n);
		return list_size * n;
	}

	// normal (faked)
	vdata[nv] = gxnormal[GXIDX (NORMAL, vat)];
	if (vdata[nv])
		nv++;
	
	if (wireframe || tex_transparent)
	{
		// color 0
		vdata[nv] = gxcol_fake[GXIDX (COL0, vat)];
		if (vdata[nv])
			nv++;
	}
	else
	{
		// color 0
		vdata[nv] = gxcol[GXIDX (COL0, vat)];
		if (vdata[nv])
			nv++;
	}

	// color 1 (faked)
	vdata[nv] = gxcol_fake[GXIDX (COL1, vat)];
	if (vdata[nv])
		nv++;

	if (HOST_TEXTURES)
	{
		if (wireframe)
		{
			// texture 0
			vdata[nv] = gxtex_fake[GXIDX (TEX0, vat)];
			if (vdata[nv])
				nv++;
		}
		else
		{
			// texture 0
			if (VCD_T0MIDX)
				vdata[nv] = gxtex_midx[GXIDX (TEX0, vat)];
			else
				vdata[nv] = gxtex[GXIDX (TEX0, vat)];
			if (vdata[nv])
				nv++;
		}

		// other textures are faked
		// texture 1
		vdata[nv] = gxtex_fake[GXIDX (TEX1, vat)];
		if (vdata[nv])
			nv++;

		// texture 2
		vdata[nv] = gxtex_fake[GXIDX (TEX2, vat)];
		if (vdata[nv])
			nv++;

		// texture 3
		vdata[nv] = gxtex_fake[GXIDX (TEX3, vat)];
		if (vdata[nv])
			nv++;

		// texture 4
		vdata[nv] = gxtex_fake[GXIDX (TEX4, vat)];
		if (vdata[nv])
			nv++;

		// texture 5
		vdata[nv] = gxtex_fake[GXIDX (TEX5, vat)];
		if (vdata[nv])
			nv++;

		// texture 6
		vdata[nv] = gxtex_fake[GXIDX (TEX6, vat)];
		if (vdata[nv])
			nv++;

		// texture 7
		vdata[nv] = gxtex_fake[GXIDX (TEX7, vat)];
		if (vdata[nv])
			nv++;

		if (!wireframe)
			gx_load_texture (0);
		else
			gx_enable_texture (0, 0);

		if (texactive[0])
		{
			DEBUG (EVENT_LOG_GX, "....  TEXTURE %.8x %dx%d",
						 texactive[0]->address, texactive[0]->width, texactive[0]->height);

			glMatrixMode (GL_TEXTURE);
	
			// dequantize
			tdq[0] = 1.0 / (1 << VAT_TEX0_SHFT (vat));
			if (VCD_T0MIDX)
			{
				glLoadIdentity ();
				if (!texactive[0]->p2)
					glScalef (TEX_SSIZE (0), TEX_TSIZE (0), 1);
			}
			else
			{
				gx_load_matrix (MIDX_TEX0, 4, 2);

				if (texactive[0]->p2)
					glScalef (tdq[0], tdq[0], tdq[0]);
				else
					glScalef (tdq[0] * TEX_SSIZE (0), tdq[0] * TEX_TSIZE (0), tdq[0]);
			}

			glMatrixMode (GL_MODELVIEW);
		}
		else
			gx_enable_texture (0, 0);
	}
	else if (texactive[0])
		gx_enable_texture (0, 0);

	// position
	i = GXIDX (POS, vat);
	psize = gxpos_size[i];
	pos = gxpos[i];

	// load geometry matrix
	vdq = 1.0 / (1 << VAT_POS_SHFT (vat));
	if (VCD_PMIDX)
	{
		pos = gxpos_midx[GXIDX (POS, vat)];
		glLoadIdentity ();
	}
	else
	{
		gx_load_matrix (MIDX_GEO, 4, 3);
		glScalef (vdq, vdq, vdq);
	}

	if (wireframe)
		glColor4f (0, 0, 0, 1);
	else
	{
		if (tex_transparent)
			glColor4f (1, 1, 1, 0.5);
		else if (sw_fullbright)
			glColor4f (1, 1, 1, 1);
		else
		{
			if ((XF (0x1009)))
			{
				color = BSWAP32 (XF_MATERIAL0);

				if (XF (0x100e) & 1)
					color |= 0xffffff00;
			
				if (XF (0x1010) & 1)
					color |= 0xff;

				glColor4ubv ((unsigned char *) &color);
			}
			else	
				glColor4f (1, 1, 1, 1);
		}
	}

	glBegin (prim);

		for (j = 0; j < n; j++)
		{
			// matrix indices
			for (i = 0; i < nm; i++)
				mem += midx[i] (mem);

			// position must be sent last
			mp = mem;
			mem += psize;

			// normals, colors and texture coords
			for (i = 0; i < nv; i++)
				mem += vdata[i] (mem);

			// position
			pos (mp);
		}

	glEnd ();

	gcube_perf_vertices (n);

	return (mem - start);
}


void gx_set_projection (void)
{
	float m[16] = {0};


	glMatrixMode (GL_PROJECTION);

	if (XF_PROJECTION_ORTHOGRAPHIC)
	{
		DEBUG (EVENT_LOG_GX, "....  setting orthographic projection matrix");

		m[ 0] = XF_PROJECTION_A;
		m[12] = XF_PROJECTION_B;
		m[ 5] = XF_PROJECTION_C;
		m[13] = XF_PROJECTION_D;
		m[10] = XF_PROJECTION_E;
		m[14] = XF_PROJECTION_F;
		m[15] = 1;

		glLoadMatrixf (m);
	}
	else
	{
		DEBUG (EVENT_LOG_GX, "....  setting perspective projection matrix");

		m[ 0] = XF_PROJECTION_A;
		m[ 8] = XF_PROJECTION_B;
		m[ 5] = XF_PROJECTION_C;
		m[ 9] = XF_PROJECTION_D;
		m[10] = XF_PROJECTION_E;
		m[14] = XF_PROJECTION_F;
		m[11] = -1;
		
		glLoadMatrixf (m);
	}
	
	glMatrixMode (GL_MODELVIEW);
}


void gx_set_lpsize (void)
{
	DEBUG (EVENT_LOG_GX, "....  SIZE point %d line %d OFFSET point %d line %d", POINT_SIZE, LINE_SIZE, POINT_TEX_OFFSET, LINE_TEX_OFFSET);

	glPointSize ((float) POINT_SIZE / 6);
	glLineWidth ((float) LINE_SIZE / 6);
}


void gx_load_tlut (void)
{
	__u32 size = ((BP_LOADTLUT1 >> 10) & 0x3ff) << 5;
	__u32 src = (BP_LOADTLUT0 & 0x1fffff) << 5;
	__u32 dst = (BP_LOADTLUT1 & 0x3ff) << 5;


	DEBUG (EVENT_LOG_GX, "....  TLUT %d MEM[%.8x] -> TMEM[%.8x]",
				 size, src | 0x80000000, dst);
	
	memcpy (TMEM_ADDRESS (dst), MEM_ADDRESS (src), size);

	gx_reload_tlut (dst);
}


void gx_set_cull_mode (void)
{
	unsigned int cull[] =
	{
		0, GL_BACK, GL_FRONT, GL_FRONT_AND_BACK,
	};


	DEBUG (EVENT_LOG_GX, "....  CULL MODE %d", CULL_MODE);

	if (wireframe)
		return;

	if (CULL_MODE == 0)
		glDisable (GL_CULL_FACE);
	else
	{
		glEnable (GL_CULL_FACE);
		glCullFace (cull[CULL_MODE]);
	}
}


#define BLEND_MODE_NONE				0
#define BLEND_MODE_BLEND			1
#define BLEND_MODE_LOGIC			2
#define BLEND_MODE_SUBSTRACT	3

void gx_set_cmode0 (void)
{
	unsigned int sfactor[] =
	{
		GL_ZERO,
		GL_ONE,
		GL_DST_ALPHA,
		GL_ONE_MINUS_DST_ALPHA,
		GL_SRC_ALPHA,
		GL_ONE_MINUS_SRC_ALPHA,
		GL_DST_COLOR,
		GL_ONE_MINUS_DST_COLOR,
	};
	unsigned int dfactor[] =
	{
		GL_ZERO,
		GL_ONE,
		GL_DST_ALPHA,
		GL_ONE_MINUS_DST_ALPHA,
		GL_SRC_ALPHA,
		GL_ONE_MINUS_SRC_ALPHA,
		GL_SRC_COLOR,
		GL_ONE_MINUS_SRC_COLOR,
	};
	unsigned int logicop[] =
	{
		GL_CLEAR,
		GL_AND,
		GL_AND_REVERSE,
		GL_COPY,
		GL_AND_INVERTED,
		GL_NOOP,
		GL_XOR,
		GL_OR,
		GL_NOR,
		GL_EQUIV,
		GL_INVERT,
		GL_OR_REVERSE,
		GL_COPY_INVERTED,
		GL_OR_INVERTED,
		GL_NAND,
		GL_SET,
	};
	unsigned int alpha_update;


	DEBUG (EVENT_LOG_GX, "....  CMODE0 BLEND %d DITHER %d UPDATE C%d/A%d DFACTOR %d SFACTOR %d LOGICOP %d",
				 CMODE_BLEND, CMODE_DITHER_ENABLED>0, CMODE_COLOR_ENABLED>0, CMODE_ALPHA_ENABLED>0,
				 CMODE_DFACTOR, CMODE_SFACTOR, CMODE_LOGICOP);

	if (CMODE_DITHER_ENABLED)
		glEnable (GL_DITHER);
	else
		glDisable (GL_DITHER);

	switch (CMODE_BLEND)
	{
		case BLEND_MODE_NONE:
			glDisable (GL_BLEND);
			glDisable (GL_COLOR_LOGIC_OP);
			break;

		case BLEND_MODE_BLEND:
			glDisable (GL_COLOR_LOGIC_OP);
			glEnable (GL_BLEND);
			glBlendFunc (sfactor[CMODE_SFACTOR], dfactor[CMODE_DFACTOR]);
			break;

		case BLEND_MODE_LOGIC:
			glDisable (GL_BLEND);
			if (!no_logic_ops)
			{
				glEnable (GL_COLOR_LOGIC_OP);
				glLogicOp (logicop[CMODE_LOGICOP]);
			}
			break;
		
		case BLEND_MODE_SUBSTRACT:
			// use extensions
			break;
	}
	
	if (CMODE_ALPHA_ENABLED)
		alpha_update = GL_TRUE;
	else
		alpha_update = GL_FALSE;

	if (CMODE_COLOR_ENABLED)
		glColorMask (GL_TRUE, GL_TRUE, GL_TRUE, alpha_update);
	else
		glColorMask (GL_FALSE, GL_FALSE, GL_FALSE, alpha_update);
}


void gx_set_zmode (void)
{
	unsigned int func[] =
	{
		GL_NEVER,
		GL_LESS,
		GL_EQUAL,
		GL_LEQUAL,
		GL_GREATER,
		GL_NOTEQUAL,
		GL_GEQUAL,
		GL_ALWAYS,
	};


	DEBUG (EVENT_LOG_GX, "....  ZMODE ENABLE %d FUNC %d UPDATE %d",
				 ZMODE_ENABLED>0, ZMODE_FUNC, ZMODE_UPDATE_ENABLED>0);

	if (ZMODE_ENABLED)
	{
		glEnable (GL_DEPTH_TEST);
		glDepthFunc (func[ZMODE_FUNC]);
	}
	else
		glDisable (GL_DEPTH_TEST);
	
	if (ZMODE_UPDATE_ENABLED)
		glDepthMask (GL_TRUE);
	else
		glDepthMask (GL_FALSE);
}


void gx_set_copy_clear_color (void)
{
	DEBUG (EVENT_LOG_GX, "....  COPY CLEAR COLOR %f %f %f %f",
				 (float) COPY_CLEAR_R / 0xff,
				 (float) COPY_CLEAR_G / 0xff,
				 (float) COPY_CLEAR_B / 0xff,
				 (float) COPY_CLEAR_A / 0xff);


	if (wireframe)
		return;

	glClearColor ((float) COPY_CLEAR_R / 0xff,
								(float) COPY_CLEAR_G / 0xff,
								(float) COPY_CLEAR_B / 0xff,
								(float) COPY_CLEAR_A / 0xff);
}


void gx_set_copy_clear_z (void)
{
	DEBUG (EVENT_LOG_GX, "....  COPY CLEAR Z %f %f %f %f",
				 (float) COPY_CLEAR_Z / 0xffffff);

	glClearDepth ((float) COPY_CLEAR_Z / 0xffffff);
}


void gx_set_scissors (void)
{
	int x, y, w, h;


	x = SCISSORS_X - 342;
	y = SCISSORS_Y - 342;
	w = SCISSORS_W - SCISSORS_X + 1;
	h = SCISSORS_H - SCISSORS_Y + 1;

	DEBUG (EVENT_LOG_GX, "....  SCISSORS %d %d %d %d", x, y, w, h);
	
	glEnable (GL_SCISSOR_TEST);
	
	h *= (float) SCREEN_HEIGHT / real_height;
	y *= (float) SCREEN_HEIGHT / real_height;
	w *= (float) SCREEN_WIDTH / real_width;
	x *= (float) SCREEN_WIDTH / real_width;
	glScissor (x, SCREEN_HEIGHT - (y + h), w, h);
}


void gx_set_viewport (void)
{
	float x, y, w, h, n, f;


	w = XF_VIEWPORT_A * 2;
	h = -XF_VIEWPORT_B * 2;
	x = XF_VIEWPORT_D - (342 + XF_VIEWPORT_A);
	y = XF_VIEWPORT_E - (342 - XF_VIEWPORT_B);
	
	f = XF_VIEWPORT_F / 16777215;
	n = (XF_VIEWPORT_F - XF_VIEWPORT_C) / 16777215;

	DEBUG (EVENT_LOG_GX, "....  VIEWPORT %.2f %.2f %.2f %.2f %.2f %.2f",
				 x, y, w, h, n, f);

	// just trying to guess the number of lines
	real_height = (FLOAT_EQ (h, 448)) ? 448 : 480;
	if (w < RVI16 (0x2070))
		real_width = RVI16 (0x2070);
	else
		real_width = w;

	h *= (float) SCREEN_HEIGHT / real_height;
	y *= (float) SCREEN_HEIGHT / real_height;
	w *= (float) SCREEN_WIDTH / real_width;
	x *= (float) SCREEN_WIDTH / real_width;
	glViewport (x, SCREEN_HEIGHT - (y + h), w, h);
	glDepthRange (n, f);
}


void gx_set_gamma (void)
{
	static const float rgb_gamma[] = { 1.0, 1.7, 2.2 };
	int gm = COPY_EXECUTE_GAMMA;


	if (last_gamma != gm)
	{
		video_set_gamma (rgb_gamma[gm]);
		last_gamma = gm;
	}
}


void gx_copy_efb (void)
{
	if (COPY_EXECUTE_TO_XFB)
		gcube_pe_refresh ();

	if (COPY_EXECUTE_CLEAR_EFB)
	{
		glViewport (0, 0, SCREEN_WIDTH, SCREEN_HEIGHT);
		glScissor (0, 0, SCREEN_WIDTH, SCREEN_HEIGHT);
		glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	}
}


GPOPCODE (INVALID)
{
	DEBUG (EVENT_EMAJOR, "..gx: invalid command %2.2x at %.8x", MEM (mem), mem | 0x80000000);

	return -1;
}


GPOPCODE (NOP)
{
	DEBUG (EVENT_LOG_GX, "....  nop");

	return 1;
}


GPOPCODE (LOAD_CP)
{
	DEBUG (EVENT_LOG_GX, "....  cp[%.2x] = %.8x", MEM (mem + 1), MEMR32 (mem + 2));

	CP (MEM (mem + 1)) = MEMR32 (mem + 2);

	return 6;
}


GPOPCODE (LOAD_XF)
{
	__u32 n = MEMR16 (mem += 1) + 1;
	__u32 a = MEMR16 (mem += 2);
	int i;


	if (EXCEEDES_LIST_BOUNDARY (mem, 5 + n*4))
		return -(5 + n*4);

	mem += 2;
	if (a < 0x800)
	{
		// load matrix
		for (i = 0; i < n; i++, mem += 4)
		{
			DEBUG (EVENT_LOG_GX, "....  xfmat[%.4x] = %.2f", a + i, MEMRF (mem));

			XFMAT (a + i) = MEMRF (mem);
		}
	}
	else
	{
		for (i = 0; i < n; i++, mem += 4)
		{
			DEBUG (EVENT_LOG_GX, "....  xf[%.4x] = %.8x", a + i, MEMR32 (mem));

			XF (a + i) = MEMR32 (mem);
		}
		
		// set matrices
		switch (a)
		{
			// projection
			case 0x1020:
				gx_set_projection ();
				break;
			
			case 0x101a:
			case 0x101b:
			case 0x101d:
			case 0x101e:
				gx_set_viewport ();
				break;
		}
	}

	return (5 + n*4);
}


// invalidate vertex cache
GPOPCODE (INVALIDATE_VCACHE)
{
	DEBUG (EVENT_LOG_GX, "....  nop");

	return 1;
}


// make it as callbacks
GPOPCODE (LOAD_BP)
{
	DEBUG (EVENT_LOG_GX, "....  bp[%.2x] = %.6x", MEM (mem + 1), MEMR32 (mem + 1) & 0x00ffffff);

	BP (MEM (mem + 1)) = MEMR32 (mem + 1);

	switch (MEM (mem + 1))
	{
		// point / line size and texture offsets
		case 0x22:
			gx_set_lpsize ();
			break;

		// TX_LOADTLUT1
		case 0x65:
			gx_load_tlut ();
			break;
	
		// TX_SETMODE0
		case 0x80:
		case 0x81:
		case 0x82:
		case 0x83:
			gx_set_texture_mode0 (MEM (mem + 1) & 3);
			break;

		case 0xa0:
		case 0xa1:
		case 0xa2:
		case 0xa3:
			gx_set_texture_mode0 ((MEM (mem + 1) & 3) + 4);
			break;

		// TX_SETMODE1
		case 0x84:
		case 0x85:
		case 0x86:
		case 0x87:
			gx_set_texture_mode1 ((MEM (mem + 1) >> 1) & 3);
			break;

		case 0xa4:
		case 0xa5:
		case 0xa6:
		case 0xa7:
			gx_set_texture_mode1 (((MEM (mem + 1) >> 1) & 3) + 4);
			break;
		
		// GEN_MODE
		case 0x00:
			gx_set_cull_mode ();
			break;
		
		// PE_ZMODE
		case 0x40:
			gx_set_zmode ();
			break;

		// PE_CMODE0
		case 0x41:
			gx_set_cmode0 ();
			break;

		case 0x4f:
		case 0x50:
			gx_set_copy_clear_color ();
			break;

		case 0x51:
			gx_set_copy_clear_z ();
			break;

		case 0x52:
//			gx_set_gamma ();
			gx_copy_efb ();
			break;

//		case 0x20:
		case 0x21:
			gx_set_scissors ();
			break;
	}

	return 5;
}


GPOPCODE (LOAD_INDEX)
{
	DEBUG (EVENT_EMAJOR, "..gx: command Load Index * not emulated yet");

	return 5;
}


GPOPCODE (CALL_DISPLAYLIST)
{
	__u32 addr = MEMR32 (mem + 1), size = MEMR32 (mem + 5);


	DEBUG (EVENT_LOG_GX, "....  call displaylist %.8x of size %.8x", addr, size);

	if (((addr & MEM_MASK) > MEM_SIZE) || (size > (CP_FIFO_END - CP_FIFO_BASE)) ||
			 (addr & 31) || (size & 31))
	{
		DEBUG (EVENT_EFATAL, "..gx: trying to parse invalid list %.8x of size %.8x",
						addr, size);

		return -9;
	}

	if (gx_parse_list (addr, size) < 0)
		return -9;
	else
	{
		DEBUG (EVENT_LOG_GX, "....  end of displaylist %.8x", addr);
		return 9;
	}
}


GPOPCODE (DRAW_QUADS)
{
	DEBUG (EVENT_LOG_GX, "....  draw quads, %d verts", MEMR16 (mem + 1));

	return 3 + gx_draw (mem + 3, GX_QUADS, MEMR16 (mem + 1), MEM (mem) & 7);
}


GPOPCODE (DRAW_TRIANGLES)
{
	DEBUG (EVENT_LOG_GX, "....  draw triangles, %d verts", MEMR16 (mem + 1));

	return 3 + gx_draw (mem + 3, GX_TRIANGLES, MEMR16 (mem + 1), MEM (mem) & 7);
}


GPOPCODE (DRAW_TRIANGLE_STRIP)
{
	DEBUG (EVENT_LOG_GX, "....  draw triangle strip, %d verts", MEMR16 (mem + 1));

	return 3 + gx_draw (mem + 3, GX_TRIANGLE_STRIP, MEMR16 (mem + 1), MEM (mem) & 7);
}


GPOPCODE (DRAW_TRIANGLE_FAN)
{
	DEBUG (EVENT_LOG_GX, "....  draw triangle fan, %d verts", MEMR16 (mem + 1));

	return 3 + gx_draw (mem + 3, GX_TRIANGLE_FAN, MEMR16 (mem + 1), MEM (mem) & 7);
}


GPOPCODE (DRAW_LINES)
{
	DEBUG (EVENT_LOG_GX, "....  draw lines, %d verts", MEMR16 (mem + 1));

	return 3 + gx_draw (mem + 3, GX_LINES, MEMR16 (mem + 1), MEM (mem) & 7);
}


GPOPCODE (DRAW_LINE_STRIP)
{
	DEBUG (EVENT_LOG_GX, "....  draw line strip, %d verts", MEMR16 (mem + 1));

	return 3 + gx_draw (mem + 3, GX_LINE_STRIP, MEMR16 (mem + 1), MEM (mem) & 7);
}


GPOPCODE (DRAW_POINTS)
{
	DEBUG (EVENT_LOG_GX, "....  draw points, %d verts", MEMR16 (mem + 1));

	return 3 + gx_draw (mem + 3, GX_POINTS, MEMR16 (mem + 1), MEM (mem) & 7);
}


int gx_switch (int sw)
{
	switch (sw)
	{
		case GX_TOGGLE_WIREFRAME:
			{
				wireframe = !wireframe;

				DEBUG (EVENT_LOG_GX, "..gx: toggle wireframe");
				if (wireframe)
				{
					glPolygonMode (GL_FRONT_AND_BACK, GL_LINE);

					glClearColor (1, 1, 1, 0);
					glDisable (GL_CULL_FACE);
				}
				else
				{
					glPolygonMode (GL_FRONT_AND_BACK, GL_FILL);
			
					gx_set_cull_mode ();
					gx_set_copy_clear_color ();
				}
			}
			return wireframe;
		
		case GX_TOGGLE_TEX_RELOAD:
			tex_reload = !tex_reload;
			return tex_reload;
		
		case GX_TOGGLE_TEX_DUMP:
			dump_textures = !dump_textures;
			return dump_textures;

		case GX_TOGGLE_FORCE_LINEAR:
			force_linear = !force_linear;
			return force_linear;

		case GX_TOGGLE_TEX_TRANSPARENT:
			tex_transparent = !tex_transparent;
			return tex_transparent;

		case GX_TOGGLE_DRAW:
			dont_draw = !dont_draw;
			return dont_draw;

		case GX_TOGGLE_NO_LOGIC_OPS:
			no_logic_ops = !no_logic_ops;
			return no_logic_ops;

		case GX_TOGGLE_FULLBRIGHT:
			sw_fullbright = !sw_fullbright;
			return sw_fullbright;

		default:
			return -1;
	}
}


void gx_reinit (void)
{
	gx_set_viewport ();
	gx_set_scissors ();
	gx_set_copy_clear_z ();
	gx_set_copy_clear_color ();
	gx_set_zmode ();
	gx_set_cmode0 ();
	gx_set_cull_mode ();
	gx_set_lpsize ();
	gx_set_projection ();
}


void gx_init (void)
{
	int i;


	memset (texbuff, 0, sizeof (texbuff));

	for (i = 0; i < 0x100; i++)
		GPOP (i) = INVALID;

	GPOP (0x00) = NOP;
	GPOP (0x08) = LOAD_CP;
	GPOP (0x10) = LOAD_XF;
	GPOP (0x20) = LOAD_INDEX;
	GPOP (0x28) = LOAD_INDEX;
	GPOP (0x30) = LOAD_INDEX;
	GPOP (0x38) = LOAD_INDEX;
	GPOP (0x40) = CALL_DISPLAYLIST;
	GPOP (0x48) = INVALIDATE_VCACHE;
	GPOP (0x61) = LOAD_BP;

	GPOPR (0x80, 0x87, DRAW_QUADS);
	GPOPR (0x90, 0x97, DRAW_TRIANGLES);
	GPOPR (0x98, 0x9f, DRAW_TRIANGLE_STRIP);
	GPOPR (0xa0, 0xa7, DRAW_TRIANGLE_FAN);
	GPOPR (0xa8, 0xaf, DRAW_LINES);
	GPOPR (0xb0, 0xb7, DRAW_LINE_STRIP);
	GPOPR (0xb8, 0xbf, DRAW_POINTS);
}
